# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# Inspired by
# https://fuchsia.googlesource.com/infra/recipes/+/336933647862a1a9718b4ca18f0a67e89c2419f8/recipe_modules/ninja/resources/ninja_wrapper.py
"""Extracts a concise error from a ninja log."""

import argparse
import logging
from pathlib import Path
import re
from typing import IO
import sys

_LOG: logging.Logger = logging.getLogger(__name__)

# Assume any of these lines could be prefixed with ANSI color codes.
_COLOR_CODES_PREFIX = r'^(?:\x1b)?(?:\[\d+m\s*)?'

_GOOGLETEST_FAILED, _GOOGLETEST_RUN, _GOOGLETEST_OK, _GOOGLETEST_DISABLED = (
    '[  FAILED  ]',
    '[ RUN      ]',
    '[       OK ]',
    '[ DISABLED ]',
)


def _remove_passing_tests(failure_lines: list[str]) -> list[str]:
    test_lines: list[str] = []
    result = []
    for line in failure_lines:
        if test_lines:
            if _GOOGLETEST_OK in line:
                test_lines = []
            elif _GOOGLETEST_FAILED in line:
                result.extend(test_lines)
                test_lines = []
                result.append(line)
            else:
                test_lines.append(line)
        elif _GOOGLETEST_RUN in line:
            test_lines.append(line)
        elif _GOOGLETEST_DISABLED in line:
            pass
        else:
            result.append(line)
    result.extend(test_lines)
    return result


def _parse_ninja(ins: IO) -> str:
    failure_lines: list[str] = []
    last_line: str = ''

    for line in ins:
        _LOG.debug('processing %r', line)
        # Trailing whitespace isn't significant, as it doesn't affect the
        # way the line shows up in the logs. However, leading whitespace may
        # be significant, especially for compiler error messages.
        line = line.rstrip()

        if failure_lines:
            _LOG.debug('inside failure block')

            if re.match(_COLOR_CODES_PREFIX + r'\[\d+/\d+\] (\S+)', line):
                _LOG.debug('next rule started, ending failure block')
                break

            if re.match(_COLOR_CODES_PREFIX + r'ninja: build stopped:.*', line):
                _LOG.debug('ninja build stopped, ending failure block')
                break
            failure_lines.append(line)

        else:
            if re.match(_COLOR_CODES_PREFIX + r'FAILED: (.*)$', line):
                _LOG.debug('starting failure block')
                failure_lines.extend([last_line, line])
            elif 'FAILED' in line:
                _LOG.debug('not a match')
        last_line = line

    failure_lines = _remove_passing_tests(failure_lines)

    # Remove "Requirement already satisfied:" lines since many of those might
    # be printed during Python installation, and they usually have no relevance
    # to the actual error.
    failure_lines = [
        x
        for x in failure_lines
        if not x.lstrip().startswith('Requirement already satisfied:')
    ]

    # Trim paths so the output is less likely to wrap. Specifically, if the
    # path has six or more segments trim it down to three.
    path_re = re.compile(
        r'(?P<prefix>\b|\s)'
        r'[-\w._]+/[-\w._]+(?:/[-\w._]+)+'
        r'(?P<core>/[-\w._]+/[-\w._]+/[-\w._]+)'
        r'(?P<suffix>\b|\s)'
    )

    def replace(m: re.Match):
        return ''.join(
            (m.group('prefix'), '[...]', m.group('core'), m.group('suffix'))
        )

    failure_lines = [path_re.sub(replace, x) for x in failure_lines]

    result: str = '\n'.join(failure_lines)
    return re.sub(r'\n+', '\n', result)


def parse_ninja_stdout(ninja_stdout: Path) -> str:
    """Extract an error summary from ninja output."""
    with ninja_stdout.open() as ins:
        return _parse_ninja(ins)


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('input', type=argparse.FileType('r'))
    parser.add_argument('output', type=argparse.FileType('w'))
    args = parser.parse_args(argv)

    for line in _parse_ninja(args.input):
        args.output.write(line)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Implementation of the custom pw-status-codes reStructuredText directive.

The main goal of this directive is to improve the consistency of the
presentation of our API references. All C++ functions and methods that return
a set of pw::Status codes use this directive.

This directive is mainly used within the Doxygen comments of C++ header files.
Usage example:

/// @returns @rst
///
/// .. pw-status-codes::
///
///    OK: The bulk read was successful.
///
///    DEADLINE_EXCEEDED: Can't acquire
///    bus access. See :ref:`x`.
///
/// @endrst

The word before the colon must be a valid pw_status code. The text after the
colon is a description of what the status code means in this particular
scenario. Descriptions may span multiple lines and can contain nested
reStructuredText.
"""
import sys
from typing import List, Tuple

from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from pw_status import Status
from sphinx.application import Sphinx


help_url = 'https://pigweed.dev/docs/style/doxygen.html#pw-status-codes'


class PwStatusCodesDirective(Directive):
    """Renders `pw-status-codes` directives."""

    has_content = True

    # The main event handler for the directive.
    # https://docutils.sourceforge.io/docs/howto/rst-directives.html
    def run(self) -> list[nodes.Node]:
        merged_data = self._merge()
        transformed_data = self._transform(merged_data)
        return self._render(transformed_data)

    def _merge(self) -> List[str]:
        """Merges the content into a format that's easier to work with.

        Docutils provides content line-by-line, with empty strings
        representing newlines:

        [
            'OK: The bulk read was successful.',
            '',
            'DEADLINE_EXCEEDED: Can't acquire',
            'bus access. See :ref:`x`.',
        ]

        _merge() consolidates each code and its accompanying description into
        a single string representing the complete key-value pair (KVP) and
        removes the empty strings:

        [
            'OK: The bulk read was successful.',
            'DEADLINE_EXCEEDED: Can't acquire bus access. See :ref:`x`.'
        ]

        We also need to handle the scenario of getting a single line of
        content:

        [
            'OK: The operation was successful.'
        ]
        """
        kvps = []  # key-value pairs
        index = 0
        kvps.append('')
        content: ViewList = self.content
        for line in content:
            # An empty line in the source content means that we're about to
            # encounter a new kvp. See the empty line after 'OK: The bulk read
            # was successful.' for an example.
            if line == '':
                # Make sure that the current kvp has content before starting
                # a new kvp.
                if kvps[index] != '':
                    index += 1
                    kvps.append('')
            else:  # The line has content we need.
                # If the current line isn't empty and our current kvp also
                # isn't empty then we're dealing with the multi-line scenario.
                # See the DEADLINE_EXCEEDED example. We just need to add back
                # the whitespace that was stripped out.
                if kvps[index] != '':
                    kvps[index] += ' '
            # All edge cases have been handled and it's now safe to always add
            # the current line to the current kvp.
            kvps[index] += line
        return kvps

    def _transform(self, items: List[str]) -> List[Tuple[str, str]]:
        """Transforms the data into normalized and annotated tuples.

        _normalize() gives us data like this:

        [
            'OK: The bulk read was successful.',
        ]

        _transform() changes it to a list of tuples where the first element
        is the status code converted into a cross-reference and the second
        element is the description.

        [
            (':c:enumerator:`OK`', 'The bulk read was successful.'),
        ]

        This function exits if any provided status code is invalid.
        """
        data = []
        valid_codes = [member.name for member in Status]
        for item in items:
            code = item.split(':', 1)[0]
            desc = item.split(':', 1)[1].strip()
            if code not in valid_codes:
                docname = self.state.document.settings.env.docname
                error = (
                    f'[error] pw-status-codes found invalid code: {code}\n'
                    f'        offending content: {item}\n'
                    f'        included in following file: {docname}.rst\n'
                    f'        help: {help_url}\n'
                )
                sys.exit(error)
            link = f':c:enumerator:`{code}`'
            data.append((link, desc))
        return data

    def _render(self, data: List[Tuple[str, str]]) -> List[nodes.Node]:
        """Renders the content as a table.

        _transform() gives us data like this:

        [
            (':c:enumerator:`OK`', 'The bulk read was successful.'),
        ]

        _render() structures it as a csv-table:

        .. csv-table:
           :header: "Code", "Description"
           :align: left

           ":c:enumerator:`OK`", "The bulk read was successful."

        And then lets Sphinx/Docutils do the rendering. This has the added
        bonus of making it possible to use inline reStructuredText elements
        like code formatting and cross-references in the descriptions.
        """
        table = [
            '.. csv-table::',
            '   :header: "Code", "Description"',
            '   :align: left',
            '',
        ]
        for item in data:
            table.append(f'   "{item[0]}", "{item[1]}"')
        container = nodes.container()
        container['classes'] = ['pw-status-codes']
        # Docutils doesn't accept normal lists; it must be a ViewList.
        # inclusive-language: disable
        # https://www.sphinx-doc.org/en/master/extdev/markupapi.html#viewlists
        # inclusive-language: enable
        self.state.nested_parse(ViewList(table), self.content_offset, container)
        return [container]


def setup(app: Sphinx) -> dict[str, bool]:
    """Registers the pw-status-codes directive in the Sphinx build system."""
    app.add_directive('pw-status-codes', PwStatusCodesDirective)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Bazel output support."""

from typing import Any

import pathlib

try:
    from pw_build_mcuxpresso.components import Project
except ImportError:
    # Load from this directory if pw_build_mcuxpresso is not available.
    from components import Project  # type: ignore


def _bazel_bool_out(name: str, val: bool, indent: int = 0) -> None:
    """Outputs boolean in Bazel format."""
    print('    ' * indent + f'{name} = "{val}",')


def _bazel_int_out(name: str, val: int, indent: int = 0) -> None:
    """Outputs integer in Bazel format."""
    print('    ' * indent + f'{name} = "{val}",')


def _bazel_str(val: Any) -> str:
    """Returns string in Bazel format with correct escaping."""
    return str(val).replace('"', r'\"').replace('$', r'\$')


def _bazel_str_out(name: str, val: Any, indent: int = 0) -> None:
    """Outputs string in Bazel format with correct escaping."""
    print('    ' * indent + f'{name} = "{_bazel_str(val)}",')


def _bazel_str_list_out(name: str, vals: list[Any], indent: int = 0) -> None:
    """Outputs list of strings in Bazel format with correct escaping."""
    if not vals:
        return

    print('    ' * indent + f'{name} = [')
    for val in vals:
        print('    ' * (indent + 1) + f'"{_bazel_str(val)}",')
    print('    ' * indent + '],')


def _bazel_path_list_out(
    name: str,
    vals: list[pathlib.Path],
    path_prefix: str | None = None,
    indent: int = 0,
) -> None:
    """Outputs list of paths in Bazel format with common prefix."""
    if path_prefix is not None:
        str_vals = [f'{path_prefix}{str(val)}' for val in vals]
    else:
        str_vals = [str(f) for f in vals]

    _bazel_str_list_out(name, sorted(set(str_vals)), indent=indent)


def bazel_output(
    project: Project,
    name: str,
    path_prefix: str | None = None,
    extra_args: dict[str, Any] | None = None,
):
    """Output Bazel target for a project with the specified components.

    Args:
        project: MCUXpresso project to output.
        name: target name to output.
        path_prefix: string prefix to prepend to all paths.
        extra_args: Dictionary of additional arguments to generated target.
    """
    print('cc_library(')
    _bazel_str_out('name', name, indent=1)
    _bazel_path_list_out(
        'srcs',
        project.sources + project.libs,
        path_prefix=path_prefix,
        indent=1,
    )
    _bazel_path_list_out(
        'hdrs', project.headers, path_prefix=path_prefix, indent=1
    )
    _bazel_str_list_out('defines', project.defines, indent=1)
    _bazel_path_list_out(
        'includes', project.include_dirs, path_prefix=path_prefix, indent=1
    )

    for arg_name, arg_value in (extra_args or {}).items():
        if isinstance(arg_value, bool):
            _bazel_bool_out(arg_name, arg_value, indent=1)
        elif isinstance(arg_value, int):
            _bazel_int_out(arg_name, arg_value, indent=1)
        elif isinstance(arg_value, str):
            _bazel_str_out(arg_name, arg_value, indent=1)
        elif isinstance(arg_value, list):
            if all(isinstance(x, str) for x in arg_value):
                _bazel_str_list_out(arg_name, arg_value, indent=1)
            else:
                raise TypeError(
                    f"Can't handle extra arg {arg_name!r}: "
                    f"a list of {type(arg_value[0])}"
                )
        else:
            raise TypeError(
                f"Can't handle extra arg {arg_name!r}: {type(arg_value)}"
            )

    print(')')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/update_bundle.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_software_update import tuf_pb2 as pw__software__update_dot_tuf__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&pw_software_update/update_bundle.proto\x12\x12pw.software_update\x1a\x1cpw_software_update/tuf.proto\"\xe5\x04\n\x0cUpdateBundle\x12L\n\x12timestamp_metadata\x18\x01 \x01(\x0b\x32+.pw.software_update.SignedTimestampMetadataH\x00\x88\x01\x01\x12J\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32*.pw.software_update.SignedSnapshotMetadataH\x01\x88\x01\x01\x12O\n\x10targets_metadata\x18\x03 \x03(\x0b\x32\x35.pw.software_update.UpdateBundle.TargetsMetadataEntry\x12M\n\x0ftarget_payloads\x18\x04 \x03(\x0b\x32\x34.pw.software_update.UpdateBundle.TargetPayloadsEntry\x12\x42\n\rroot_metadata\x18\x05 \x01(\x0b\x32&.pw.software_update.SignedRootMetadataH\x02\x88\x01\x01\x1a\x61\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).pw.software_update.SignedTargetsMetadata:\x02\x38\x01\x1a\x35\n\x13TargetPayloadsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\x15\n\x13_timestamp_metadataB\x14\n\x12_snapshot_metadataB\x10\n\x0e_root_metadata\"\xe2\x01\n\x08Manifest\x12K\n\x10targets_metadata\x18\x01 \x03(\x0b\x32\x31.pw.software_update.Manifest.TargetsMetadataEntry\x12\x1a\n\ruser_manifest\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x1a[\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.pw.software_update.TargetsMetadata:\x02\x38\x01\x42\x10\n\x0e_user_manifestB\"\n\x1e\x64\x65v.pigweed.pw_software_updateP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_software_update.update_bundle_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\036dev.pigweed.pw_software_updateP\001'
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._options = None
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._options = None
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_options = b'8\001'
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._options = None
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_options = b'8\001'
  _globals['_UPDATEBUNDLE']._serialized_start=93
  _globals['_UPDATEBUNDLE']._serialized_end=706
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_start=491
  _globals['_UPDATEBUNDLE_TARGETSMETADATAENTRY']._serialized_end=588
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_start=590
  _globals['_UPDATEBUNDLE_TARGETPAYLOADSENTRY']._serialized_end=643
  _globals['_MANIFEST']._serialized_start=709
  _globals['_MANIFEST']._serialized_end=935
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_start=826
  _globals['_MANIFEST_TARGETSMETADATAENTRY']._serialized_end=917
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""A subprocess wrapper that enables injection of externally-provided tools."""

import abc
from pathlib import Path
from typing import Iterable

import subprocess


class ToolRunner(abc.ABC):
    """A callable interface that runs the requested tool as a subprocess.

    This class is used to support subprocess-like semantics while allowing
    injection of wrappers that enable testing, finer granularity identifying
    where tools fail, and stricter control of which binaries are called.

    By default, all subprocess output is captured.
    """

    def __call__(
        self,
        tool: str,
        args: Iterable[str | Path],
        stdout: int | None = subprocess.PIPE,
        stderr: int | None = subprocess.PIPE,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        """Calls ``tool`` with the provided ``args``.

        ``**kwargs`` are forwarded to the underlying ``subprocess.run()``
        for the requested tool.

        By default, all subprocess output is captured.

        Returns:
            The ``subprocess.CompletedProcess`` result of running the requested
            tool.
        """
        additional_kwargs = set(self._custom_args())
        allowed_kwargs = {
            key: value
            for key, value in kwargs.items()
            if not key.startswith('pw_') or key in additional_kwargs
        }

        return self._run_tool(
            tool,
            args,
            stderr=stderr,
            stdout=stdout,
            **allowed_kwargs,
        )

    @staticmethod
    def _custom_args() -> Iterable[str]:
        """List of additional keyword arguments accepted by this tool.

        By default, all kwargs passed into a tool are forwarded to
        ``subprocess.run()``. However, some tools have extra arguments custom
        to them, which are not valid for ``subprocess.run()``. Tools requiring
        these custom args should override this method, listing the arguments
        they accept.

        To make filtering custom arguments possible, they must be prefixed
        with  ``pw_``.
        """
        return []

    @abc.abstractmethod
    def _run_tool(
        self, tool: str, args, **kwargs
    ) -> subprocess.CompletedProcess:
        """Implements the subprocess runner logic.

        Calls ``tool`` with the provided ``args``. ``**kwargs`` not listed in
        ``_custom_args`` are forwarded to the underlying ``subprocess.run()``
        for the requested tool.

        Returns:
            The ``subprocess.CompletedProcess`` result of running the requested
            tool.
        """


class BasicSubprocessRunner(ToolRunner):
    """A simple ToolRunner that calls subprocess.run()."""

    def _run_tool(
        self, tool: str, args, **kwargs
    ) -> subprocess.CompletedProcess:
        return subprocess.run([tool, *args], **kwargs)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Infrastructure used by the user interface or specific emulators."""

import io
import json
import logging
import os
import re
import socket
import subprocess
import sys
import time

from abc import ABC, abstractmethod
from importlib import import_module
from pathlib import Path
from typing import Any, Type

import psutil  # type: ignore

from pw_emu.pigweed_emulators import pigweed_emulators
from pw_env_setup.config_file import load as pw_config_load
from pw_env_setup.config_file import path as pw_config_path
from serial import Serial


_LAUNCHER_LOG = logging.getLogger('pw_qemu.core.launcher')


def _stop_process(pid: int) -> None:
    """Gracefully stops a running process."""

    try:
        proc = psutil.Process(pid)
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except psutil.TimeoutExpired:
            proc.kill()
    except psutil.NoSuchProcess:
        pass


def _get_class(name: str) -> type:
    """Returns a class from a full qualified class name
    (e.g. "package.module.Class").

    """
    try:
        module_path, class_name = name.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError):
        raise ImportError(name)


class Error(Exception):
    """Generic pw_emu exception."""


class ConfigError(Error):
    """Exception raised for configuration errors."""

    def __init__(self, config: Path | None, err: str) -> None:
        msg = f'{config}: {err}\n'
        try:
            if config:
                with open(config, 'r') as file:
                    msg += json.dumps(json.load(file), indent=4)
        except (OSError, json.decoder.JSONDecodeError):
            pass
        super().__init__(msg)


class AlreadyRunning(Error):
    """Exception raised if an emulator process is already running."""

    def __init__(self, wdir: Path) -> None:
        super().__init__(f'{wdir}: emulator already started')


class NotRunning(Error):
    """Exception raised if an emulator process is not running."""

    def __init__(self, wdir: Path) -> None:
        super().__init__(f'{wdir}: emulator not started')


class InvalidEmulator(Error):
    """Exception raised if an different backend is running."""

    def __init__(self, emu: str) -> None:
        super().__init__(f'invalid emulator `{emu}`')


class InvalidTarget(Error):
    """Exception raised if the target is invalid."""

    def __init__(self, config: Path, emu: str | None, target: str) -> None:
        emu_str = f'for `{emu}`' if emu else ''
        super().__init__(f'{config}: invalid target `{target}` {emu_str}')


class InvalidChannelName(Error):
    """Exception raised if a channel name is invalid."""

    def __init__(self, name: str, target: str, valid: str) -> None:
        msg = f"""
        `{name}` is not a valid device name for {target}`
        try: {valid}
        """
        super().__init__(msg)


class InvalidChannelType(Error):
    """Exception raised if a channel type is invalid."""

    def __init__(self, name: str) -> None:
        super().__init__(f'`{name}` is not a valid channel type')


class WrongEmulator(Error):
    """Exception raised if a different backend is running."""

    def __init__(self, exp: str, found: str) -> None:
        super().__init__(f'wrong emulator: expected `{exp}, found {found}`')


class RunError(Error):
    """Exception raised when a command failed to run."""

    def __init__(self, proc: str, msg: str) -> None:
        super().__init__(f'error running `{proc}`: {msg}')
        self.proc = proc
        self.msg = msg


class InvalidPropertyPath(Error):
    """Exception raised for an invalid property path."""

    def __init__(self, path: str) -> None:
        super().__init__(f'invalid property path `{path}`')


class InvalidProperty(Error):
    """Exception raised for an invalid property path."""

    def __init__(self, path: str, name: str) -> None:
        super().__init__(f'invalid property `{name}` at `{path}`')


class HandlesError(Error):
    """Exception raised if the load of an emulator handle fails."""

    def __init__(self, msg: str) -> None:
        super().__init__(f'error loading handles: {msg}')


class Handles:
    """Running emulator handles."""

    class Channel:
        def __init__(self, chan_type: str):
            self.type = chan_type

    class PtyChannel(Channel):
        def __init__(self, path: str):
            super().__init__('pty')
            self.path = path

    class TcpChannel(Channel):
        def __init__(self, host: str, port: int):
            super().__init__('tcp')
            self.host = host
            self.port = port

    class Proc:
        def __init__(self, pid: int):
            self.pid = pid

    @staticmethod
    def _ser_obj(obj) -> Any:
        if isinstance(obj, dict):
            data = {}
            for key, val in obj.items():
                data[key] = Handles._ser_obj(val)
            return data
        if hasattr(obj, "__iter__") and not isinstance(obj, str):
            return [Handles._ser_obj(item) for item in obj]
        if hasattr(obj, "__dict__"):
            return Handles._ser_obj(obj.__dict__)
        return obj

    def _serialize(self):
        return Handles._ser_obj(self)

    def save(self, wdir: Path) -> None:
        """Saves handles to the given working directory."""

        with open(os.path.join(wdir, 'handles.json'), 'w') as file:
            json.dump(self._serialize(), file)

    @staticmethod
    def load(wdir: Path):
        try:
            with open(os.path.join(wdir, 'handles.json'), 'r') as file:
                data = json.load(file)
        except (KeyError, OSError, json.decoder.JSONDecodeError):
            raise NotRunning(wdir)

        handles = Handles(data['emu'], data['config'])
        handles.set_target(data['target'])
        gdb_cmd = data.get('gdb_cmd')
        if gdb_cmd:
            handles.set_gdb_cmd(gdb_cmd)
        for name, chan in data['channels'].items():
            chan_type = chan['type']
            if chan_type == 'tcp':
                handles.add_channel_tcp(name, chan['host'], chan['port'])
            elif chan_type == 'pty':
                handles.add_channel_pty(name, chan['path'])
            else:
                raise InvalidChannelType(chan_type)
        for name, proc in data['procs'].items():
            handles.add_proc(name, proc['pid'])
        return handles

    def __init__(self, emu: str, config: str) -> None:
        self.emu = emu
        self.config = config
        self.gdb_cmd: list[str] = []
        self.target = ''
        self.channels: dict[str, Handles.Channel] = {}
        self.procs: dict[str, Handles.Proc] = {}

    def add_channel_tcp(self, name: str, host: str, port: int) -> None:
        """Adds a TCP channel."""

        self.channels[name] = self.TcpChannel(host, port)

    def add_channel_pty(self, name: str, path: str) -> None:
        """Adds a pty channel."""

        self.channels[name] = self.PtyChannel(path)

    def add_proc(self, name: str, pid: int) -> None:
        """Adds a process ID."""

        self.procs[name] = self.Proc(pid)

    def set_target(self, target: str) -> None:
        """Sets the target."""

        self.target = target

    def set_gdb_cmd(self, cmd: list[str]) -> None:
        """Sets the ``gdb`` command."""

        self.gdb_cmd = cmd.copy()


def _stop_processes(handles: Handles, wdir: Path) -> None:
    """Stops all processes for a (partially) running emulator instance.

    Remove pid files as well.
    """

    for _, proc in handles.procs.items():
        _stop_process(proc.pid)
        path = os.path.join(wdir, f'{proc}.pid')
        if os.path.exists(path):
            os.unlink(path)


class Config:
    """Get and validate options from the configuration file."""

    def __init__(
        self,
        config_path: Path | None = None,
        target: str | None = None,
        emu: str | None = None,
    ) -> None:
        """Loads the emulator configuration.

        If no configuration file path is given, the root project
        configuration is used.

        This method set ups the generic configuration (e.g. ``gdb``).

        It loads emulator target files and gathers them under the ``targets``
        key for each emulator backend. The ``targets`` settings in the
        configuration file takes precedence over the loaded target files.

        """
        try:
            if config_path:
                with open(config_path, 'r') as file:
                    config = json.load(file)['pw']['pw_emu']
            else:
                config_path = pw_config_path()
                config = pw_config_load()['pw']['pw_emu']
        except KeyError:
            raise ConfigError(config_path, 'missing `pw_emu` configuration')

        if not config_path:
            raise ConfigError(None, 'unable to deterine config path')

        if config.get('target_files'):
            tmp = {}
            for path in config['target_files']:
                if not os.path.isabs(path):
                    path = os.path.join(os.path.dirname(config_path), path)
                with open(path, 'r') as file:
                    tmp.update(json.load(file).get('targets'))
            if config.get('targets'):
                tmp.update(config['targets'])
            config['targets'] = tmp

        self.path = config_path
        self._config = {'emulators': pigweed_emulators}
        self._config.update(config)
        self._emu = emu
        self._target = target

    def set_target(self, target: str) -> None:
        """Sets the current target.

        The current target is used by the
        :py:meth:`pw_emu.core.Config.get_target` method.

        """

        self._target = target
        try:
            self.get(['targets', target], optional=False, entry_type=dict)
        except ConfigError:
            raise InvalidTarget(self.path, self._emu, self._target)

    def get_targets(self) -> list[str]:
        return list(self.get(['targets'], entry_type=dict).keys())

    def _subst(self, string: str) -> str:
        """Substitutes $pw_<subst_type>{arg} statements."""

        match = re.search(r'\$pw_([^{]+){([^}]+)}', string)
        if not match:
            return string

        subst_type = match.group(1)
        arg = match.group(2)

        if subst_type == 'env':
            value = os.environ.get(arg)
            if value is None:
                msg = f'Environment variable `{arg}` not set'
                raise ConfigError(self.path, msg)
            return string.replace(f'$pw_{subst_type}{{{arg}}}', value)

        raise ConfigError(self.path, f'Invalid substitution type: {subst_type}')

    def _subst_list(self, items: list[Any]) -> list[Any]:
        new_list = []
        for item in items:
            if isinstance(item, str):
                new_list.append(self._subst(item))
            else:
                new_list.append(item)
        return new_list

    def get(
        self,
        keys: list[str],
        optional: bool = True,
        entry_type: Type | None = None,
    ) -> Any:
        """Gets a config entry.

        ``keys`` identifies the config entry, e.g.
        ``['targets', 'test-target']`` looks in the config dictionary for
        ``['targets']['test-target']``.

        If the option is not found and optional is ``True`` it returns ``None``
        if ``entry_type`` is ``None`` or a new (empty) object of type
        ``entry_type``.

        If the option is not found and ``optional`` is ``False`` it raises
        ``ConfigError``.

        If ``entry_type`` is not ``None`` it checks the option to be of
        that type. If it is not it will raise ``ConfigError``.

        """

        keys_str = ': '.join(keys)
        entry: dict[str, Any] | None = self._config

        for key in keys:
            if not isinstance(entry, dict):
                if optional:
                    if entry_type:
                        return entry_type()
                    return None
                raise ConfigError(self.path, f'{keys_str}: not found')
            entry = entry.get(key)

        if entry is None:
            if optional:
                if entry_type:
                    return entry_type()
                return None
            raise ConfigError(self.path, f'{keys_str}: not found')

        if entry_type and not isinstance(entry, entry_type):
            msg = f'{keys_str}: expected entry of type `{entry_type}`'
            raise ConfigError(self.path, msg)

        if isinstance(entry, str):
            entry = self._subst(entry)
        elif isinstance(entry, list):
            entry = self._subst_list(entry)

        return entry

    def get_target(
        self,
        keys: list[str],
        optional: bool = True,
        entry_type: Type | None = None,
    ) -> Any:
        """Gets a config option starting at ``['targets'][target]``."""

        if not self._target:
            raise Error('target not set')
        return self.get(['targets', self._target] + keys, optional, entry_type)

    def get_emu(
        self,
        keys: list[str],
        optional: bool = True,
        entry_type: Type | None = None,
    ) -> Any:
        """Gets a config option starting at ``[emu]``."""

        if not self._emu:
            raise Error('emu not set')
        return self.get([self._emu] + keys, optional, entry_type)

    def get_target_emu(
        self,
        keys: list[str],
        optional: bool = True,
        entry_type: Type | None = None,
    ) -> Any:
        """Gets a config option starting at ``['targets'][target][emu]``."""

        if not self._emu or not self._target:
            raise Error('emu or target not set')
        return self.get(
            ['targets', self._target, self._emu] + keys, optional, entry_type
        )


class Connector(ABC):
    """Interface between a running emulator and the user-visible APIs."""

    def __init__(self, wdir: Path) -> None:
        self._wdir = wdir
        self._handles = Handles.load(wdir)
        self._channels = self._handles.channels
        self._target = self._handles.target

    @staticmethod
    def get(wdir: Path) -> Any:
        """Returns a connector instance for a given emulator type."""
        handles = Handles.load(wdir)
        config = Config(handles.config)
        emu = handles.emu
        try:
            name = config.get(['emulators', emu, 'connector'])
            cls = _get_class(name)
        except (ConfigError, ImportError):
            raise InvalidEmulator(emu)
        return cls(wdir)

    def get_emu(self) -> str:
        """Returns the emulator type."""

        return self._handles.emu

    def get_gdb_cmd(self) -> list[str]:
        """Returns the configured ``gdb`` command."""
        return self._handles.gdb_cmd

    def get_config_path(self) -> Path:
        """Returns the configuration path."""

        return self._handles.config

    def get_procs(self) -> dict[str, Handles.Proc]:
        """Returns the running processes indexed by the process name."""

        return self._handles.procs

    def get_channel_type(self, name: str) -> str:
        """Returns the channel type."""

        try:
            return self._channels[name].type
        except KeyError:
            channels = ' '.join(self._channels.keys())
            raise InvalidChannelName(name, self._target, channels)

    def get_channel_path(self, name: str) -> str:
        """Returns the channel path. Raises ``InvalidChannelType`` if this
        is not a ``pty`` channel.

        """

        try:
            if self._channels[name].type != 'pty':
                raise InvalidChannelType(self._channels[name].type)
            return self._channels[name].path
        except KeyError:
            raise InvalidChannelName(name, self._target, self._channels.keys())

    def get_channel_addr(self, name: str) -> tuple:
        """Returns a pair of ``(host, port)`` for the channel. Raises
        ``InvalidChannelType`` if this is not a ``tcp`` channel.

        """

        try:
            if self._channels[name].type != 'tcp':
                raise InvalidChannelType(self._channels[name].type)
            return (self._channels[name].host, self._channels[name].port)
        except KeyError:
            raise InvalidChannelName(name, self._target, self._channels.keys())

    def get_channel_stream(
        self,
        name: str,
        timeout: float | None = None,
    ) -> io.RawIOBase:
        """Returns a file object for a given host-exposed device.

        If ``timeout`` is ``None`` then reads and writes are blocking. If
        ``timeout`` is ``0`` the stream is operating in non-blocking
        mode. Otherwise reads and writes will timeout after the given
        value.

        """

        chan_type = self.get_channel_type(name)
        if chan_type == 'tcp':
            host, port = self.get_channel_addr(name)
            if ':' in host:
                sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            else:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((host, port))
            sock.settimeout(timeout)
            ret = sock.makefile('rwb', buffering=0)
            sock.close()
            return ret
        if chan_type == 'pty':
            ser = Serial(self.get_channel_path(name))
            ser.timeout = timeout
            return ser
        raise InvalidChannelType(chan_type)

    def get_channels(self) -> list[str]:
        return self._handles.channels.keys()

    def get_logs(self) -> str:
        """Returns the emulator logs."""

        log_path = self._wdir / f'{self._handles.emu}.log'
        return log_path.read_text()

    def stop(self) -> None:
        """Stops the emulator."""

        _stop_processes(self._handles, self._wdir)

        try:
            os.unlink(os.path.join(self._wdir, 'handles.json'))
        except OSError:
            pass

    def proc_running(self, proc: str) -> bool:
        try:
            return psutil.pid_exists(self._handles.procs[proc].pid)
        except (NotRunning, KeyError):
            return False

    def running(self) -> bool:
        """Checks if the main emulator process is already running."""

        try:
            return psutil.pid_exists(self._handles.procs[self._handles.emu].pid)
        except (NotRunning, KeyError):
            return False

    @abstractmethod
    def reset(self) -> None:
        """Performs a software reset."""

    @abstractmethod
    def cont(self) -> None:
        """Resumes the emulator's execution."""

    @abstractmethod
    def list_properties(self, path: str) -> list[Any]:
        """Returns the property list for an emulator object."""

    @abstractmethod
    def set_property(self, path: str, prop: str, value: str) -> None:
        """Sets the value of an emulator's object property."""

    @abstractmethod
    def get_property(self, path: str, prop: str) -> Any:
        """Returns the value of an emulator's object property."""


class Launcher(ABC):
    """Starts an emulator based on the target and configuration file."""

    def __init__(
        self,
        emu: str,
        config_path: Path | None = None,
    ) -> None:
        """Initializes a ``Launcher`` instance."""

        self._wdir: Path | None = None
        """Working directory"""

        self._emu = emu
        """Emulator type (e.g. "qemu", "renode")."""

        self._target: str | None = None
        """Target, initialized to None and set with _prep_start."""

        self._config = Config(config_path, emu=emu)
        """Global, emulator and  target configuration."""

        self._handles = Handles(self._emu, str(self._config.path))
        """Handles for processes, channels, etc."""

        gdb_cmd = self._config.get(['gdb'], entry_type=list)
        if gdb_cmd:
            self._handles.set_gdb_cmd(gdb_cmd)

    @staticmethod
    def get(emu: str, config_path: Path | None = None) -> Any:
        """Returns a launcher for a given emulator type."""
        config = Config(config_path)
        try:
            name = config.get(['emulators', emu, 'launcher'])
            cls = _get_class(name)
        except (ConfigError, ImportError):
            raise InvalidEmulator(str(emu))
        return cls(config_path)

    @abstractmethod
    def _pre_start(
        self,
        target: str,
        file: Path | None = None,
        pause: bool = False,
        debug: bool = False,
        args: str | None = None,
    ) -> list[str]:
        """Pre-start work, returns command to start the emulator.

        The target and emulator configuration can be accessed through
        :py:attr:`pw_emu.core.Launcher._config` with
        :py:meth:`pw_emu.core.Config.get`,
        :py:meth:`pw_emu.core.Config.get_target`,
        :py:meth:`pw_emu.core.Config.get_emu`,
        :py:meth:`pw_emu.core.Config.get_target_emu`.
        """

    @abstractmethod
    def _post_start(self) -> None:
        """Post-start work, finalize emulator handles.

        Perform any post-start emulator initialization and finalize the emulator
        handles information.

        Typically an internal monitor channel is used to inquire information
        about the the configured channels (e.g. TCP ports, pty paths) and
        :py:attr:`pw_emu.core.Launcher._handles` is updated via
        :py:meth:`pw_emu.core.Handles.add_channel_tcp`,
        :py:meth:`pw_emu.core.Handles.add_channel_pty`, etc.

        """

    @abstractmethod
    def _get_connector(self, wdir: Path) -> Connector:
        """Gets a connector for this emulator type."""

    def _path(self, name: Path | str) -> Path:
        """Returns the full path for a given emulator file."""
        if self._wdir is None:
            raise Error('internal error')
        return Path(os.path.join(self._wdir, name))

    def _subst_channel(self, subst_type: str, arg: str, string: str) -> str:
        """Substitutes $pw_emu_channel_{func}{arg} statements."""

        try:
            chan = self._handles.channels[arg]
        except KeyError:
            return string

        if subst_type == 'channel_port':
            if not isinstance(chan, Handles.TcpChannel):
                return string
            return str(chan.port)

        if subst_type == 'channel_host':
            if not isinstance(chan, Handles.TcpChannel):
                return string
            return chan.host

        if subst_type == 'channel_path':
            if not isinstance(chan, Handles.PtyChannel):
                return string
            return chan.path

        return string

    def _subst(self, string: str) -> str:
        """Substitutes $pw_emu_<subst_type>{arg} statements."""

        match = re.search(r'\$pw_emu_([^{]+){([^}]+)}', string)
        if not match:
            return string

        subst_type = match.group(1)
        arg = match.group(2)

        if subst_type == 'wdir':
            if self._wdir:
                return os.path.join(self._wdir, arg)
            return string

        if 'channel_' in subst_type:
            return self._subst_channel(subst_type, arg, string)

        return string

    # pylint: disable=protected-access
    # use os._exit after fork instead of os.exit
    def _daemonize(
        self,
        name: str,
        cmd: list[str],
    ) -> None:
        """Daemonize process for UNIX hosts."""

        if sys.platform == 'win32':
            raise Error('_daemonize not supported on win32')

        # pylint: disable=no-member
        # avoid pylint false positive on win32
        pid = os.fork()
        if pid < 0:
            raise RunError(name, f'fork failed: {pid}')
        if pid > 0:
            return

        path: Path = Path('/dev/null')
        fd = os.open(path, os.O_RDONLY)
        os.dup2(fd, sys.stdin.fileno())
        os.close(fd)

        path = self._path(f'{name}.log')
        fd = os.open(path, os.O_WRONLY | os.O_TRUNC | os.O_CREAT)
        os.dup2(fd, sys.stdout.fileno())
        os.dup2(fd, sys.stderr.fileno())
        os.close(fd)

        os.setsid()

        if os.fork() > 0:
            os._exit(0)

        try:
            # Make the pid file create and pid write operations atomic to avoid
            # races with readers.
            with open(self._path(f'{name}.pid.tmp'), 'w') as file:
                file.write(f'{os.getpid()}')
            os.rename(self._path(f'{name}.pid.tmp'), self._path(f'{name}.pid'))
            os.execvp(cmd[0], cmd)
        finally:
            os._exit(1)

    def _start_proc(
        self,
        name: str,
        cmd: list[str],
        foreground: bool = False,
    ) -> subprocess.Popen | None:
        """Run the main emulator process.

        The process pid is stored and can later be accessed by its name to
        terminate it when the emulator is stopped.

        If foreground is True the process run in the foreground and a
        subprocess.Popen object is returned. Otherwise the process is started in
        the background and None is returned.

        When running in the background stdin is redirected to the NULL device
        and stdout and stderr are redirected to a file named <name>.log which is
        stored in the emulator's instance working directory.

        """
        for idx, item in enumerate(cmd):
            cmd[idx] = self._subst(item)

        pid_file_path = self._path(f'{name}.pid')
        if os.path.exists(pid_file_path):
            os.unlink(pid_file_path)

        if foreground:
            proc = subprocess.Popen(cmd)
            self._handles.add_proc(name, proc.pid)
            with open(pid_file_path, 'w') as file:
                file.write(f'{proc.pid}')
            return proc

        if sys.platform == 'win32':
            file = open(self._path(f'{name}.log'), 'w')
            proc = subprocess.Popen(
                cmd,
                stdin=subprocess.DEVNULL,
                stdout=file,
                stderr=file,
                creationflags=subprocess.DETACHED_PROCESS,
            )
            file.close()
            with open(pid_file_path, 'w') as file:
                file.write(f'{proc.pid}')
            self._handles.add_proc(name, proc.pid)
            # avoids resource warnings due to not calling wait which
            # we don't want to do since we've started the process in
            # the background
            proc.returncode = 0
        else:
            self._daemonize(name, cmd)

        # wait for the pid file to avoid double start race conditions
        timeout = time.monotonic() + 30
        while not os.path.exists(self._path(f'{name}.pid')):
            time.sleep(0.1)
            if time.monotonic() > timeout:
                break
        if not os.path.exists(self._path(f'{name}.pid')):
            raise RunError(name, 'pid file timeout')
        try:
            with open(pid_file_path, 'r') as file:
                pid = int(file.readline())
                self._handles.add_proc(name, pid)
        except (OSError, ValueError) as err:
            raise RunError(name, str(err))

        return None

    def _stop_procs(self):
        """Stop all registered processes."""

        for name, proc in self._handles.procs.items():
            _stop_process(proc.pid)
            if os.path.exists(self._path(f'{name}.pid')):
                os.unlink(self._path(f'{name}.pid'))

    def _start_procs(self, procs_list: str) -> None:
        """Start additional processes besides the main emulator one."""

        procs = self._config.get_target([procs_list], entry_type=dict)
        for name, cmd in procs.items():
            self._start_proc(name, cmd)

    def start(
        self,
        wdir: Path,
        target: str,
        file: Path | None = None,
        pause: bool = False,
        debug: bool = False,
        foreground: bool = False,
        args: str | None = None,
    ) -> Connector:
        """Starts the emulator for the given target.

        If ``file`` is set the emulator loads that file before starting.

        If ``pause`` is ``True`` the emulator gets paused.

        If ``debug`` is ``True`` the emulator runs in the foreground with
        debug output enabled. This is useful for seeing errors, traces, etc.

        If ``foreground`` is ``True`` the emulator is run in the foreground
        otherwise it is started in daemon mode. This is useful when there is
        another process controlling the emulator's life cycle, e.g. cuttlefish.

        ``args`` are passed directly to the emulator.

        """

        try:
            handles = Handles.load(wdir)
            if psutil.pid_exists(handles.procs[handles.emu].pid):
                raise AlreadyRunning(wdir)
        except NotRunning:
            pass

        self._wdir = wdir
        self._target = target
        self._config.set_target(target)
        self._handles.set_target(target)
        gdb_cmd = self._config.get_target(['gdb'], entry_type=list)
        if gdb_cmd:
            self._handles.set_gdb_cmd(gdb_cmd)
        os.makedirs(wdir, mode=0o700, exist_ok=True)

        cmd = self._pre_start(
            target=target, file=file, pause=pause, debug=debug, args=args
        )

        if debug:
            foreground = True
            _LAUNCHER_LOG.setLevel(logging.DEBUG)

        _LAUNCHER_LOG.debug('starting emulator with command: %s', ' '.join(cmd))

        try:
            self._start_procs('pre-start-cmds')
            proc = self._start_proc(self._emu, cmd, foreground)
            self._start_procs('post-start-cmds')
        except RunError as err:
            self._stop_procs()
            raise err

        try:
            self._post_start()
        except RunError as err:
            self._handles.save(wdir)
            connector = self._get_connector(self._wdir)
            if not connector.running():
                msg = err.msg + '; dumping logs:\n' + connector.get_logs()
                raise RunError(err.proc, msg)
            raise err
        self._handles.save(wdir)

        if proc:
            proc.wait()
            self._stop_procs()

        return self._get_connector(self._wdir)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ListRequest(google.protobuf.message.Message):
    """A ListRequest has the following properties:

     - A request with an empty `path` string is valid and will list the contents
       at the "root" directory.
     - Only exact path matches will be resolved (i.e. no prefix matching).
     - Paths should be treated as case-sensitive.
     - The provided path must be absolute. If no matching path is found, a
       NOT_FOUND error is raised.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    path: builtins.str
    def __init__(
        self,
        *,
        path: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["path", b"path"]) -> None: ...

global___ListRequest = ListRequest

@typing_extensions.final
class DeleteRequest(google.protobuf.message.Message):
    """A DeleteRequest has the following properties:

     - Only exact path matches will be resolved (i.e. no prefix matching).
     - Paths should be treated as case-sensitive.
     - Deletion of directories is implementation-defined, and may be
       disallowed and return an UNIMPLEMENTED error.
     - The provided path must be absolute. If no matching path is found, a
       NOT_FOUND error is raised.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    path: builtins.str
    def __init__(
        self,
        *,
        path: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["path", b"path"]) -> None: ...

global___DeleteRequest = DeleteRequest

@typing_extensions.final
class Path(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Permissions:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _PermissionsEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Path._Permissions.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NONE: Path._Permissions.ValueType  # 0
        READ: Path._Permissions.ValueType  # 1
        WRITE: Path._Permissions.ValueType  # 2
        READ_AND_WRITE: Path._Permissions.ValueType  # 3

    class Permissions(_Permissions, metaclass=_PermissionsEnumTypeWrapper):
        """This enum is a bitmask of permissions:
        Bit 0: readable.
        Bit 1: writable.
        """

    NONE: Path.Permissions.ValueType  # 0
    READ: Path.Permissions.ValueType  # 1
    WRITE: Path.Permissions.ValueType  # 2
    READ_AND_WRITE: Path.Permissions.ValueType  # 3

    PATH_FIELD_NUMBER: builtins.int
    PERMISSIONS_FIELD_NUMBER: builtins.int
    SIZE_BYTES_FIELD_NUMBER: builtins.int
    FILE_ID_FIELD_NUMBER: builtins.int
    path: builtins.str
    """A path to a file/directory. This path is relative to the requested path
    to reduce transmission of redundant information.
    """
    permissions: global___Path.Permissions.ValueType
    """Permitted operations on this path."""
    size_bytes: builtins.int
    """The size of the file at this path."""
    file_id: builtins.int
    """A globally-unique transfer ID for this file path (e.g. for use with
    pw_transfer's RPC service). It is implementation defined whether a file's
    file ID is stable or ephemeral.
    """
    def __init__(
        self,
        *,
        path: builtins.str = ...,
        permissions: global___Path.Permissions.ValueType | None = ...,
        size_bytes: builtins.int | None = ...,
        file_id: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "_permissions", b"_permissions", "_size_bytes", b"_size_bytes", "file_id", b"file_id", "permissions", b"permissions", "size_bytes", b"size_bytes"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_file_id", b"_file_id", "_permissions", b"_permissions", "_size_bytes", b"_size_bytes", "file_id", b"file_id", "path", b"path", "permissions", b"permissions", "size_bytes", b"size_bytes"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_file_id", b"_file_id"]) -> typing_extensions.Literal["file_id"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_permissions", b"_permissions"]) -> typing_extensions.Literal["permissions"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_size_bytes", b"_size_bytes"]) -> typing_extensions.Literal["size_bytes"] | None: ...

global___Path = Path

@typing_extensions.final
class ListResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATHS_FIELD_NUMBER: builtins.int
    @property
    def paths(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Path]:
        """Each returned Path's path name is always relative to the requested path to
        reduce transmission of redundant information.
        """
    def __init__(
        self,
        *,
        paths: collections.abc.Iterable[global___Path] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["paths", b"paths"]) -> None: ...

global___ListResponse = ListResponse



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of the NXP mcuxpresso SDK."""

from pathlib import Path
from typing import Sequence

import pw_package.git_repo
import pw_package.package_manager


class McuxpressoSdk(pw_package.package_manager.Package):
    """Install and check status of the NXP mcuxpresso SDK."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='mcuxpresso', **kwargs)
        self._evkmimxrt595 = pw_package.git_repo.GitRepo(
            name='evkmimxrt595',
            url=(
                'https://pigweed-internal.googlesource.com/'
                'third_party/vendor/nxp/evkmimxrt595'
            ),
            commit='003f1764ad3e03b9162bfc02918659c4bcb3eb01',
        )

    def install(self, path: Path) -> None:
        self._evkmimxrt595.install(path)

    def info(self, path: Path) -> Sequence[str]:
        return (
            f'{self.name} installed in: {path}',
            "Enable by running 'gn args out' and adding this line:",
            f'  dir_pw_third_party_mcuxpresso = "{path}"',
        )


pw_package.package_manager.register(McuxpressoSdk)



#!/usr/bin/env python3

# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Restructured Text Formatting."""

import argparse
from dataclasses import dataclass, field
import difflib
from functools import cached_property
from pathlib import Path
import textwrap
from typing import Iterable

from pw_cli.diff import colorize_diff

TAB_WIDTH = 8  # Number of spaces to use for \t replacement
CODE_BLOCK_INDENTATION = 3


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        '--diff',
        action='store_true',
        help='Print a diff of formatting changes.',
    )
    parser.add_argument(
        '-i',
        '--in-place',
        action='store_true',
        help='Replace existing file with the reformatted copy.',
    )
    parser.add_argument(
        'rst_files',
        nargs='+',
        default=[],
        type=Path,
        help='Paths to rst files.',
    )

    return parser.parse_args()


def _indent_amount(line: str) -> int:
    return len(line) - len(line.lstrip())


def _reindent(input_text: str, amount: int) -> Iterable[str]:
    for line in textwrap.dedent(input_text).splitlines():
        if len(line.strip()) == 0:
            yield '\n'
            continue
        yield f'{" " * amount}{line}\n'


def _fix_whitespace(line: str) -> str:
    return line.rstrip().replace('\t', ' ' * TAB_WIDTH) + '\n'


@dataclass
class CodeBlock:
    """Store a single code block."""

    directive_lineno: int
    directive_line: str
    first_line_indent: int | None = None
    end_lineno: int | None = None
    option_lines: list[str] = field(default_factory=list)
    code_lines: list[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self._blank_line_after_options_found = False

    def finished(self) -> bool:
        return self.end_lineno is not None

    def append_line(self, index: int, line: str) -> None:
        """Process a line for this code block."""
        # Check if outside the code block (indentation is less):
        if (
            self.first_line_indent is not None
            and len(line.strip()) > 0
            and _indent_amount(line) < self.first_line_indent
        ):
            # Code block ended
            self.end_lineno = index
            return

        # If first line indent hasn't been found
        if self.first_line_indent is None:
            # Check if the first word is a directive option.
            # E.g. :caption:
            line_words = line.split()
            if (
                line_words
                and line_words[0].startswith(':')
                and line_words[0].endswith(':')
                # In case the first word starts with two colons '::'
                and len(line_words[0]) > 2
            ):
                self.option_lines.append(line.rstrip())
                return

            # Step 1: Check for a blank line
            if len(line.strip()) == 0:
                if (
                    self.option_lines
                    and not self._blank_line_after_options_found
                ):
                    self._blank_line_after_options_found = True
                return

            # Step 2: Check for a line that is a continuation of a previous
            # option.
            if self.option_lines and not self._blank_line_after_options_found:
                self.option_lines.append(line.rstrip())
                return

            # Step 3: Check a line with content.
            if len(line.strip()) > 0:
                # Line is not a directive and not blank: it is content.
                # Flag the end of the options
                self._blank_line_after_options_found = True

            # Set the content indentation amount.
            self.first_line_indent = _indent_amount(line)

        # Save this line as code.
        self.code_lines.append(self._clean_up_line(line))

    def _clean_up_line(self, line: str) -> str:
        line = line.rstrip()
        if not self._keep_codeblock_tabs:
            line = line.replace('\t', ' ' * TAB_WIDTH)
        return line

    @cached_property
    def _keep_codeblock_tabs(self) -> bool:
        """True if tabs should NOT be replaced; keep for 'none' or 'go'."""
        return 'none' in self.directive_line or 'go' in self.directive_line

    @cached_property
    def directive_indent_amount(self) -> int:
        return _indent_amount(self.directive_line)

    def options_block_lines(self) -> Iterable[str]:
        yield from _reindent(
            input_text='\n'.join(self.option_lines),
            amount=self.directive_indent_amount + CODE_BLOCK_INDENTATION,
        )

    def code_block_lines(self) -> Iterable[str]:
        yield from _reindent(
            input_text='\n'.join(self.code_lines),
            amount=self.directive_indent_amount + CODE_BLOCK_INDENTATION,
        )

    def lines(self) -> Iterable[str]:
        """Yields the code block directives's lines."""
        yield self.directive_line
        if self.option_lines:
            yield from self.options_block_lines()
        yield '\n'
        yield from self.code_block_lines()
        yield '\n'

    def __repr__(self) -> str:
        return ''.join(self.lines())


def _parse_and_format_rst(in_text: str) -> Iterable[str]:
    """Reindents code blocks to 3 spaces and fixes whitespace."""
    current_block: CodeBlock | None = None
    for index, line in enumerate(in_text.splitlines(keepends=True)):
        # If a code block is active, process this line.
        if current_block:
            current_block.append_line(index, line)
            if current_block.finished():
                yield from current_block.lines()
                # This line wasn't part of the code block, process as normal.
                yield _fix_whitespace(line)
                # Erase this code_block variable
                current_block = None
        # Check for new code block start
        elif line.lstrip().startswith(('.. code-block::', '.. code::')):
            current_block = CodeBlock(
                directive_lineno=index,
                # Change `.. code::` to Sphinx's `.. code-block::`.
                directive_line=line.replace('code::', 'code-block::'),
            )
            continue
        else:
            yield _fix_whitespace(line)
    # If the document ends with a code block it may still need to be written.
    if current_block is not None:
        yield from current_block.lines()


def reformat_rst(
    file_name: Path,
    diff: bool = False,
    in_place: bool = False,
    suppress_stdout: bool = False,
) -> list[str]:
    """Reformats an rst file and returns a list of diff lines."""
    in_text = file_name.read_text()
    out_lines = list(_parse_and_format_rst(in_text))

    # Remove blank lines from the end of the output, if any.
    while out_lines and not out_lines[-1].strip():
        out_lines.pop()

    # Add a trailing \n if needed.
    if out_lines and not out_lines[-1].endswith('\n'):
        out_lines[-1] += '\n'

    result_diff = list(
        difflib.unified_diff(
            in_text.splitlines(True),
            out_lines,
            f'{file_name}  (original)',
            f'{file_name}  (reformatted)',
        )
    )
    if diff and result_diff:
        if not suppress_stdout:
            print(''.join(colorize_diff(result_diff)))

    if in_place:
        file_name.write_text(''.join(out_lines))

    return result_diff


def rst_format_main(
    rst_files: list[Path],
    diff: bool = False,
    in_place: bool = False,
) -> None:
    for rst_file in rst_files:
        reformat_rst(rst_file, diff, in_place)


if __name__ == '__main__':
    rst_format_main(**vars(_parse_args()))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Bandwidth Monitor Toolbar and Tracker."""
from __future__ import annotations

from contextvars import ContextVar
import logging
import textwrap
from typing import TYPE_CHECKING

from prompt_toolkit.layout import WindowAlign

from pw_console.plugin_mixin import PluginMixin
from pw_console.widgets import ToolbarButton, WindowPaneToolbar
from pw_console.widgets.event_count_history import EventCountHistory

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer

_LOG = logging.getLogger('pw_console.serial_debug_logger')


def _log_hex_strings(data: bytes, prefix=''):
    """Create alinged hex number and character view log messages."""
    # Make a list of 2 character hex number strings.
    hex_numbers = textwrap.wrap(data.hex(), 2)

    hex_chars = [
        ('<' + str(b.to_bytes(1, byteorder='big')) + '>')
        .replace("<b'\\x", '', 1)  # Remove b'\x from the beginning
        .replace("<b'", '', 1)  # Remove b' from the beginning
        .replace("'>", '', 1)  # Remove ' from the end
        .rjust(2)
        for b in data
    ]

    # Replace non-printable bytes with dots.
    for i, num in enumerate(hex_numbers):
        if num == hex_chars[i]:
            hex_chars[i] = '..'

    hex_numbers_msg = ' '.join(hex_numbers)
    hex_chars_msg = ' '.join(hex_chars)

    _LOG.debug(
        '%s%s',
        prefix,
        hex_numbers_msg,
        extra=dict(
            extra_metadata_fields={
                'msg': hex_numbers_msg,
                'view': 'hex',
            }
        ),
    )
    _LOG.debug(
        '%s%s',
        prefix,
        hex_chars_msg,
        extra=dict(
            extra_metadata_fields={
                'msg': hex_chars_msg,
                'view': 'chars',
            }
        ),
    )


BANDWIDTH_HISTORY_CONTEXTVAR = ContextVar(
    'pw_console_bandwidth_history',
    default={
        'total': EventCountHistory(interval=3),
        'read': EventCountHistory(interval=3),
        'write': EventCountHistory(interval=3),
    },
)


class SerialBandwidthTracker:
    """Tracks and logs the data read and written by a serial tranport."""

    def __init__(self):
        self.pw_bps_history = BANDWIDTH_HISTORY_CONTEXTVAR.get()

    def track_read_data(self, data: bytes) -> None:
        """Tracks and logs data read."""
        self.pw_bps_history['read'].log(len(data))
        self.pw_bps_history['total'].log(len(data))

        if len(data) > 0:
            prefix = 'Read %2d B: ' % len(data)
            _LOG.debug(
                '%s%s',
                prefix,
                data,
                extra=dict(
                    extra_metadata_fields={
                        'mode': 'Read',
                        'bytes': len(data),
                        'view': 'bytes',
                        'msg': str(data),
                    }
                ),
            )
            _log_hex_strings(data, prefix=prefix)

            # Print individual lines
            for line in data.decode(
                encoding='utf-8', errors='ignore'
            ).splitlines():
                _LOG.debug(
                    '%s',
                    line,
                    extra=dict(
                        extra_metadata_fields={
                            'msg': line,
                            'view': 'lines',
                        }
                    ),
                )

    def track_write_data(self, data: ReadableBuffer) -> None:
        """Tracks and logs data to be written."""
        if isinstance(data, bytes) and len(data) > 0:
            self.pw_bps_history['write'].log(len(data))
            self.pw_bps_history['total'].log(len(data))

            prefix = 'Write %2d B: ' % len(data)
            _LOG.debug(
                '%s%s',
                prefix,
                data,
                extra=dict(
                    extra_metadata_fields={
                        'mode': 'Write',
                        'bytes': len(data),
                        'view': 'bytes',
                        'msg': str(data),
                    }
                ),
            )
            _log_hex_strings(data, prefix=prefix)

            # Print individual lines
            for line in data.decode(
                encoding='utf-8', errors='ignore'
            ).splitlines():
                _LOG.debug(
                    '%s',
                    line,
                    extra=dict(
                        extra_metadata_fields={
                            'msg': line,
                            'view': 'lines',
                        }
                    ),
                )


class BandwidthToolbar(WindowPaneToolbar, PluginMixin):
    """Toolbar for displaying bandwidth history."""

    TOOLBAR_HEIGHT = 1

    def _update_toolbar_text(self):
        """Format toolbar text.

        This queries pyserial_wrapper's EventCountHistory context var to
        retrieve the byte count history for read, write and totals."""
        tokens = []
        self.plugin_logger.debug('BandwidthToolbar _update_toolbar_text')

        for count_name, events in self.history.items():
            tokens.extend(
                [
                    ('', '  '),
                    (
                        'class:theme-bg-active class:theme-fg-active',
                        ' {}: '.format(count_name.title()),
                    ),
                    (
                        'class:theme-bg-active class:theme-fg-cyan',
                        '{:.3f} '.format(events.last_count()),
                    ),
                    (
                        'class:theme-bg-active class:theme-fg-orange',
                        '{} '.format(events.display_unit_title),
                    ),
                ]
            )
            if count_name == 'total':
                tokens.append(
                    ('class:theme-fg-cyan', '{}'.format(events.sparkline()))
                )

        self.formatted_text = tokens

    def get_left_text_tokens(self):
        """Formatted text to display on the far left side."""
        return self.formatted_text

    def get_right_text_tokens(self):
        """Formatted text to display on the far right side."""
        return [('class:theme-fg-blue', 'Serial Bandwidth Usage ')]

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, center_section_align=WindowAlign.RIGHT, **kwargs
        )

        self.history = BANDWIDTH_HISTORY_CONTEXTVAR.get()
        self.show_toolbar = True
        self.formatted_text = []

        # Buttons for display in the center
        self.add_button(
            ToolbarButton(
                description='Refresh', mouse_handler=self._update_toolbar_text
            )
        )

        # Set plugin options
        self.background_task_update_count: int = 0
        self.plugin_init(
            plugin_callback=self._background_task,
            plugin_callback_frequency=1.0,
            plugin_logger_name='pw_console_bandwidth_toolbar',
        )

    def _background_task(self) -> bool:
        self.background_task_update_count += 1
        self._update_toolbar_text()
        self.plugin_logger.debug(
            'BandwidthToolbar Scheduled Update: #%s',
            self.background_task_update_count,
        )
        return True



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class SnapshotTraceInfo(google.protobuf.message.Message):
    """This message overlays the pw.snapshot.Snapshot proto. It's valid to encode
    this message to the same sink that a Snapshot proto is being written to.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRACE_DATA_FIELD_NUMBER: builtins.int
    trace_data: builtins.bytes
    """If this field is populated, it is expected to contain a de-ringed tokenized
    trace buffer.
    """
    def __init__(
        self,
        *,
        trace_data: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["trace_data", b"trace_data"]) -> None: ...

global___SnapshotTraceInfo = SnapshotTraceInfo



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed renode frontend."""

import socket
import time
import xmlrpc.client

from pathlib import Path
from typing import Any

from pw_emu.core import (
    Connector,
    Error,
    Handles,
    InvalidChannelType,
    Launcher,
    RunError,
    WrongEmulator,
)


class RenodeRobotError(Error):
    """Exception for Renode robot errors."""

    def __init__(self, err: str):
        super().__init__(err)


class RenodeLauncher(Launcher):
    """Start a new renode process for a given target and config file."""

    def __init__(self, config_path: Path | None = None):
        super().__init__('renode', config_path)
        self._start_cmd: list[str] = []

    @staticmethod
    def _allocate_port() -> int:
        """Allocate renode ports.

        This is inherently racy but renode currently does not have proper
        support for dynamic ports. It accecept 0 as a port and the OS allocates
        a dynamic port but there is no API to retrive the port.

        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('localhost', 0))
        port = sock.getsockname()[1]
        sock.close()

        return port

    def _pre_start(
        self,
        target: str,
        file: Path | None = None,
        pause: bool = False,
        debug: bool = False,
        args: str | None = None,
    ) -> list[str]:
        renode = self._config.get_target_emu(['executable'])
        if not renode:
            renode = self._config.get_emu(['executable'], optional=False)

        self._start_cmd.extend([f'{renode}', '--disable-xwt'])
        port = self._allocate_port()
        self._start_cmd.extend(['--robot-server-port', str(port)])
        self._handles.add_channel_tcp('robot', 'localhost', port)

        machine = self._config.get_target_emu(['machine'], optional=False)
        self._start_cmd.extend(['--execute', f'mach add "{target}"'])

        self._start_cmd.extend(
            ['--execute', f'machine LoadPlatformDescription @{machine}']
        )

        terms = self._config.get_target_emu(
            ['channels', 'terminals'], entry_type=dict
        )
        for name in terms.keys():
            port = self._allocate_port()
            dev_path = self._config.get_target_emu(
                ['channels', 'terminals', name, 'device-path'],
                optional=False,
                entry_type=str,
            )
            term_type = self._config.get_target_emu(
                ['channels', 'terminals', name, 'type'],
                entry_type=str,
            )
            if not term_type:
                term_type = self._config.get_emu(
                    ['channels', 'terminals', 'type'],
                    entry_type=str,
                )
            if not term_type:
                term_type = 'tcp'

            cmd = 'emulation '
            if term_type == 'tcp':
                cmd += f'CreateServerSocketTerminal {port} "{name}" false'
                self._handles.add_channel_tcp(name, 'localhost', port)
            elif term_type == 'pty':
                path = self._path(name)
                cmd += f'CreateUartPtyTerminal "{name}" "{path}"'
                self._handles.add_channel_pty(name, str(path))
            else:
                raise InvalidChannelType(term_type)

            self._start_cmd.extend(['--execute', cmd])
            self._start_cmd.extend(
                ['--execute', f'connector Connect {dev_path} {name}']
            )

        port = self._allocate_port()
        self._start_cmd.extend(['--execute', f'machine StartGdbServer {port}'])
        self._handles.add_channel_tcp('gdb', 'localhost', port)

        if file:
            self._start_cmd.extend(['--execute', f'sysbus LoadELF @{file}'])

        if not pause:
            self._start_cmd.extend(['--execute', 'start'])

        return self._start_cmd

    def _post_start(self) -> None:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        robot = self._handles.channels['gdb']
        assert isinstance(robot, Handles.TcpChannel)

        # renode is slow to start especially during host load
        deadline = time.monotonic() + 120
        connected = False
        err = None
        while time.monotonic() < deadline:
            try:
                sock.connect((robot.host, robot.port))
                connected = True
                break
            except OSError as exc:
                err = exc
            time.sleep(1)

        if not connected:
            msg = 'failed to connect to robot channel'
            msg += f'({robot.host}:{robot.port}): {err}'
            raise RunError('renode', msg)

        sock.close()

    def _get_connector(self, wdir: Path) -> Connector:
        return RenodeConnector(wdir)


class RenodeConnector(Connector):
    """renode implementation for the emulator specific connector methods."""

    def __init__(self, wdir: Path) -> None:
        super().__init__(wdir)
        if self.get_emu() != 'renode':
            raise WrongEmulator('renode', self.get_emu())
        robot = self._handles.channels['robot']
        host = robot.host
        port = robot.port
        self._proxy = xmlrpc.client.ServerProxy(f'http://{host}:{port}/')

    def _request(self, cmd: str, args: list[str]) -> Any:
        """Send a request using the robot interface.

        Using the robot interface is not ideal since it is designed
        for testing. However, it is more robust than the ANSI colored,
        echoed, log mixed, telnet interface.

        """

        resp = self._proxy.run_keyword(cmd, args)
        if not isinstance(resp, dict):
            raise RenodeRobotError('expected dictionary in response')
        if resp['status'] != 'PASS':
            raise RenodeRobotError(resp['error'])
        if resp.get('return'):
            return resp['return']
        return None

    def reset(self) -> None:
        self._request('ResetEmulation', [])

    def cont(self) -> None:
        self._request('StartEmulation', [])

    def list_properties(self, path: str) -> list[Any]:
        return self._request('ExecuteCommand', [f'{path}'])

    def get_property(self, path: str, prop: str) -> Any:
        return self._request('ExecuteCommand', [f'{path} {prop}'])

    def set_property(self, path: str, prop: str, value: Any) -> None:
        return self._request('ExecuteCommand', [f'{path} {prop} {value}'])



#!/usr/bin/env python3

# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Runs the local presubmit checks for the Pigweed repository."""

import argparse
import json
import logging
import os
from pathlib import Path
import platform
import re
import shutil
import subprocess
import sys
from typing import Callable, Iterable, Sequence, TextIO

from pw_cli.plural import plural
from pw_cli.file_filter import FileFilter
import pw_package.pigweed_packages
from pw_presubmit import (
    build,
    cli,
    cpp_checks,
    format_code,
    git_repo,
    gitmodules,
    inclusive_language,
    javascript_checks,
    json_check,
    keep_sorted,
    module_owners,
    npm_presubmit,
    owners_checks,
    python_checks,
    shell_checks,
    source_in_build,
    todo_check,
)
from pw_presubmit.presubmit import (
    Programs,
    call,
    filter_paths,
)
from pw_presubmit.presubmit_context import (
    PresubmitContext,
    PresubmitFailure,
)
from pw_presubmit.tools import log_run
from pw_presubmit.install_hook import install_git_hook

_LOG = logging.getLogger(__name__)

pw_package.pigweed_packages.initialize()

# Trigger builds if files with these extensions change.
_BUILD_FILE_FILTER = FileFilter(
    suffix=(
        *format_code.C_FORMAT.extensions,
        '.cfg',
        '.py',
        '.rst',
        '.gn',
        '.gni',
        '.emb',
    )
)

_OPTIMIZATION_LEVELS = 'debug', 'size_optimized', 'speed_optimized'


def _at_all_optimization_levels(target):
    for level in _OPTIMIZATION_LEVELS:
        yield f'{target}_{level}'


class PigweedGnGenNinja(build.GnGenNinja):
    """Add Pigweed-specific defaults to GnGenNinja."""

    def add_default_gn_args(self, args):
        """Add project-specific default GN args to 'args'."""
        args['pw_C_OPTIMIZATION_LEVELS'] = ('debug',)


def build_bazel(*args, **kwargs) -> None:
    build.bazel(
        *args, use_remote_cache=True, strict_module_lockfile=True, **kwargs
    )


#
# Build presubmit checks
#
gn_all = PigweedGnGenNinja(
    name='gn_all',
    path_filter=_BUILD_FILE_FILTER,
    gn_args=dict(pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS),
    ninja_targets=('all',),
)


def gn_clang_build(ctx: PresubmitContext):
    """Checks all compile targets that rely on LLVM tooling."""
    build_targets = [
        *_at_all_optimization_levels('host_clang'),
        'cpp20_compatibility',
        'asan',
        'tsan',
        'ubsan',
        'runtime_sanitizers',
        # TODO: b/234876100 - msan will not work until the C++ standard library
        # included in the sysroot has a variant built with msan.
    ]

    # clang-tidy doesn't run on Windows.
    if sys.platform != 'win32':
        build_targets.append('static_analysis')

    # QEMU doesn't run on Windows.
    if sys.platform != 'win32':
        # TODO: b/244604080 - For the pw::InlineString tests, qemu_clang_debug
        #     and qemu_clang_speed_optimized produce a binary too large for the
        #     QEMU target's 256KB flash. Restore debug and speed optimized
        #     builds when this is fixed.
        build_targets.append('qemu_clang_size_optimized')

    # TODO: b/240982565 - SocketStream currently requires Linux.
    if sys.platform.startswith('linux'):
        build_targets.append('integration_tests')

    build.gn_gen(ctx, pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS)
    build.ninja(ctx, *build_targets)
    build.gn_check(ctx)


_HOST_COMPILER = 'gcc' if sys.platform == 'win32' else 'clang'


@filter_paths(file_filter=_BUILD_FILE_FILTER)
def gn_quick_build_check(ctx: PresubmitContext):
    """Checks the state of the GN build by running gn gen and gn check."""
    build.gn_gen(ctx)


def _gn_main_build_check_targets() -> Sequence[str]:
    build_targets = [
        'check_modules',
        *_at_all_optimization_levels('stm32f429i'),
        *_at_all_optimization_levels(f'host_{_HOST_COMPILER}'),
        'python.tests',
        'python.lint',
        'pigweed_pypi_distribution',
    ]

    return build_targets


def _gn_platform_build_check_targets() -> Sequence[str]:
    build_targets = []

    # TODO: b/315998985 - Add docs back to Mac ARM build.
    if sys.platform != 'darwin' or platform.machine() != 'arm64':
        build_targets.append('docs')

    # C headers seem to be missing when building with pw_minimal_cpp_stdlib, so
    # skip it on Windows.
    if sys.platform != 'win32':
        build_targets.append('build_with_pw_minimal_cpp_stdlib')

    # TODO: b/234645359 - Re-enable on Windows when compatibility tests build.
    if sys.platform != 'win32':
        build_targets.append('cpp20_compatibility')

    # clang-tidy doesn't run on Windows.
    if sys.platform != 'win32':
        build_targets.append('static_analysis')

    # QEMU doesn't run on Windows.
    if sys.platform != 'win32':
        # TODO: b/244604080 - For the pw::InlineString tests, qemu_*_debug
        #     and qemu_*_speed_optimized produce a binary too large for the
        #     QEMU target's 256KB flash. Restore debug and speed optimized
        #     builds when this is fixed.
        build_targets.append('qemu_gcc_size_optimized')
        build_targets.append('qemu_clang_size_optimized')

    # TODO: b/240982565 - SocketStream currently requires Linux.
    if sys.platform.startswith('linux'):
        build_targets.append('integration_tests')

    # TODO: b/269354373 - clang is not supported on windows yet
    if sys.platform != 'win32':
        build_targets.append('host_clang_debug_dynamic_allocation')

    return build_targets


def _gn_combined_build_check_targets() -> Sequence[str]:
    return [
        *_gn_main_build_check_targets(),
        *_gn_platform_build_check_targets(),
    ]


gn_main_build_check = PigweedGnGenNinja(
    name='gn_main_build_check',
    doc='Run most host.',
    path_filter=_BUILD_FILE_FILTER,
    gn_args=dict(
        pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS,
        pw_BUILD_BROKEN_GROUPS=True,  # Enable to fully test the GN build
    ),
    ninja_targets=_gn_main_build_check_targets(),
)

gn_platform_build_check = PigweedGnGenNinja(
    name='gn_platform_build_check',
    doc='Run any host platform-specific tests.',
    path_filter=_BUILD_FILE_FILTER,
    gn_args=dict(
        pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS,
        pw_BUILD_BROKEN_GROUPS=True,  # Enable to fully test the GN build
    ),
    ninja_targets=_gn_platform_build_check_targets(),
)

gn_combined_build_check = PigweedGnGenNinja(
    name='gn_combined_build_check',
    doc='Run most host and device (QEMU) tests.',
    path_filter=_BUILD_FILE_FILTER,
    gn_args=dict(
        pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS,
        pw_BUILD_BROKEN_GROUPS=True,  # Enable to fully test the GN build
    ),
    ninja_targets=_gn_combined_build_check_targets(),
)

coverage = PigweedGnGenNinja(
    name='coverage',
    doc='Run coverage for the host build.',
    path_filter=_BUILD_FILE_FILTER,
    ninja_targets=('coverage',),
    coverage_options=build.CoverageOptions(
        common=build.CommonCoverageOptions(
            target_bucket_project='pigweed',
            target_bucket_root='gs://ng3-metrics/ng3-pigweed-coverage',
            trace_type='LLVM',
            owner='pigweed-infra@google.com',
            bug_component='503634',
        ),
        codesearch=(
            build.CodeSearchCoverageOptions(
                host='pigweed-internal',
                project='codesearch',
                add_prefix='pigweed',
                ref='refs/heads/main',
                source='infra:main',
            ),
        ),
        gerrit=build.GerritCoverageOptions(
            project='pigweed/pigweed',
        ),
    ),
)


@filter_paths(file_filter=_BUILD_FILE_FILTER)
def gn_arm_build(ctx: PresubmitContext):
    build.gn_gen(ctx, pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS)
    build.ninja(ctx, *_at_all_optimization_levels('stm32f429i'))
    build.gn_check(ctx)


stm32f429i = PigweedGnGenNinja(
    name='stm32f429i',
    path_filter=_BUILD_FILE_FILTER,
    gn_args={
        'pw_use_test_server': True,
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_contexts=(
        lambda ctx: build.test_server(
            'stm32f429i_disc1_test_server',
            ctx.output_dir,
        ),
    ),
    ninja_targets=_at_all_optimization_levels('stm32f429i'),
)

gn_crypto_mbedtls_build = PigweedGnGenNinja(
    name='gn_crypto_mbedtls_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('mbedtls',),
    gn_args={
        'dir_pw_third_party_mbedtls': lambda ctx: '"{}"'.format(
            ctx.package_root / 'mbedtls'
        ),
        'pw_crypto_SHA256_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_crypto:sha256_mbedtls_v3'
        ),
        'pw_crypto_ECDSA_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_crypto:ecdsa_mbedtls_v3'
        ),
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=(
        *_at_all_optimization_levels(f'host_{_HOST_COMPILER}'),
        # TODO: b/240982565 - SocketStream currently requires Linux.
        *(('integration_tests',) if sys.platform.startswith('linux') else ()),
    ),
)

gn_crypto_micro_ecc_build = PigweedGnGenNinja(
    name='gn_crypto_micro_ecc_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('micro-ecc',),
    gn_args={
        'dir_pw_third_party_micro_ecc': lambda ctx: '"{}"'.format(
            ctx.package_root / 'micro-ecc'
        ),
        'pw_crypto_ECDSA_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_crypto:ecdsa_uecc'
        ),
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=(
        *_at_all_optimization_levels(f'host_{_HOST_COMPILER}'),
        # TODO: b/240982565 - SocketStream currently requires Linux.
        *(('integration_tests',) if sys.platform.startswith('linux') else ()),
    ),
)

gn_teensy_build = PigweedGnGenNinja(
    name='gn_teensy_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('teensy',),
    gn_args={
        'pw_arduino_build_CORE_PATH': lambda ctx: '"{}"'.format(
            str(ctx.package_root)
        ),
        'pw_arduino_build_CORE_NAME': 'teensy',
        'pw_arduino_build_PACKAGE_NAME': 'avr/1.58.1',
        'pw_arduino_build_BOARD': 'teensy40',
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=_at_all_optimization_levels('arduino'),
)

gn_pico_build = PigweedGnGenNinja(
    name='gn_pico_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('pico_sdk', 'freertos'),
    gn_args={
        'dir_pw_third_party_freertos': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'freertos')
        ),
        'PICO_SRC_DIR': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'pico_sdk')
        ),
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=('pi_pico',),
)

gn_mimxrt595_build = PigweedGnGenNinja(
    name='gn_mimxrt595_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('mcuxpresso',),
    gn_args={
        'dir_pw_third_party_mcuxpresso': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'mcuxpresso')
        ),
        'pw_target_mimxrt595_evk_MANIFEST': '$dir_pw_third_party_mcuxpresso'
        + '/EVK-MIMXRT595_manifest_v3_13.xml',
        'pw_third_party_mcuxpresso_SDK': '//targets/mimxrt595_evk:sample_sdk',
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=('mimxrt595'),
)

gn_mimxrt595_freertos_build = PigweedGnGenNinja(
    name='gn_mimxrt595_freertos_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('freertos', 'mcuxpresso'),
    gn_args={
        'dir_pw_third_party_freertos': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'freertos')
        ),
        'dir_pw_third_party_mcuxpresso': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'mcuxpresso')
        ),
        'pw_target_mimxrt595_evk_freertos_MANIFEST': '{}/{}'.format(
            "$dir_pw_third_party_mcuxpresso", "EVK-MIMXRT595_manifest_v3_13.xml"
        ),
        'pw_third_party_mcuxpresso_SDK': '//targets/mimxrt595_evk_freertos:sdk',
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=('mimxrt595_freertos'),
)

gn_software_update_build = PigweedGnGenNinja(
    name='gn_software_update_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('nanopb', 'protobuf', 'mbedtls', 'micro-ecc'),
    gn_args={
        'dir_pw_third_party_protobuf': lambda ctx: '"{}"'.format(
            ctx.package_root / 'protobuf'
        ),
        'dir_pw_third_party_nanopb': lambda ctx: '"{}"'.format(
            ctx.package_root / 'nanopb'
        ),
        'dir_pw_third_party_micro_ecc': lambda ctx: '"{}"'.format(
            ctx.package_root / 'micro-ecc'
        ),
        'pw_crypto_ECDSA_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_crypto:ecdsa_uecc'
        ),
        'dir_pw_third_party_mbedtls': lambda ctx: '"{}"'.format(
            ctx.package_root / 'mbedtls'
        ),
        'pw_crypto_SHA256_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_crypto:sha256_mbedtls_v3'
        ),
        'pw_C_OPTIMIZATION_LEVELS': _OPTIMIZATION_LEVELS,
    },
    ninja_targets=_at_all_optimization_levels('host_clang'),
)

gn_pw_system_demo_build = PigweedGnGenNinja(
    name='gn_pw_system_demo_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('freertos', 'nanopb', 'stm32cube_f4', 'pico_sdk'),
    gn_args={
        'dir_pw_third_party_freertos': lambda ctx: '"{}"'.format(
            ctx.package_root / 'freertos'
        ),
        'dir_pw_third_party_nanopb': lambda ctx: '"{}"'.format(
            ctx.package_root / 'nanopb'
        ),
        'dir_pw_third_party_stm32cube_f4': lambda ctx: '"{}"'.format(
            ctx.package_root / 'stm32cube_f4'
        ),
        'PICO_SRC_DIR': lambda ctx: '"{}"'.format(
            str(ctx.package_root / 'pico_sdk')
        ),
    },
    ninja_targets=('pw_system_demo',),
)

gn_chre_googletest_nanopb_sapphire_build = PigweedGnGenNinja(
    name='gn_chre_googletest_nanopb_sapphire_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('boringssl', 'chre', 'emboss', 'googletest', 'nanopb'),
    gn_args=dict(
        dir_pw_third_party_chre=lambda ctx: '"{}"'.format(
            ctx.package_root / 'chre'
        ),
        dir_pw_third_party_nanopb=lambda ctx: '"{}"'.format(
            ctx.package_root / 'nanopb'
        ),
        dir_pw_third_party_googletest=lambda ctx: '"{}"'.format(
            ctx.package_root / 'googletest'
        ),
        dir_pw_third_party_emboss=lambda ctx: '"{}"'.format(
            ctx.package_root / 'emboss'
        ),
        dir_pw_third_party_boringssl=lambda ctx: '"{}"'.format(
            ctx.package_root / 'boringssl'
        ),
        pw_unit_test_MAIN=lambda ctx: '"{}"'.format(
            ctx.root / 'third_party/googletest:gmock_main'
        ),
        pw_unit_test_BACKEND=lambda ctx: '"{}"'.format(
            ctx.root / 'pw_unit_test:googletest'
        ),
        pw_function_CONFIG=lambda ctx: '"{}"'.format(
            ctx.root / 'pw_function:enable_dynamic_allocation'
        ),
        pw_bluetooth_sapphire_ENABLED=True,
        pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS,
    ),
    ninja_targets=(
        *_at_all_optimization_levels(f'host_{_HOST_COMPILER}'),
        *_at_all_optimization_levels('stm32f429i'),
    ),
)

gn_fuzz_build = PigweedGnGenNinja(
    name='gn_fuzz_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('abseil-cpp', 'fuzztest', 'googletest', 're2'),
    gn_args={
        'dir_pw_third_party_abseil_cpp': lambda ctx: '"{}"'.format(
            ctx.package_root / 'abseil-cpp'
        ),
        'dir_pw_third_party_fuzztest': lambda ctx: '"{}"'.format(
            ctx.package_root / 'fuzztest'
        ),
        'dir_pw_third_party_googletest': lambda ctx: '"{}"'.format(
            ctx.package_root / 'googletest'
        ),
        'dir_pw_third_party_re2': lambda ctx: '"{}"'.format(
            ctx.package_root / 're2'
        ),
        'pw_unit_test_MAIN': lambda ctx: '"{}"'.format(
            ctx.root / 'third_party/googletest:gmock_main'
        ),
        'pw_unit_test_BACKEND': lambda ctx: '"{}"'.format(
            ctx.root / 'pw_unit_test:googletest'
        ),
    },
    ninja_targets=('fuzzers',),
    ninja_contexts=(
        lambda ctx: build.modified_env(
            FUZZTEST_PRNG_SEED=build.fuzztest_prng_seed(ctx),
        ),
    ),
)

oss_fuzz_build = PigweedGnGenNinja(
    name='oss_fuzz_build',
    path_filter=_BUILD_FILE_FILTER,
    packages=('abseil-cpp', 'fuzztest', 'googletest', 're2'),
    gn_args={
        'dir_pw_third_party_abseil_cpp': lambda ctx: '"{}"'.format(
            ctx.package_root / 'abseil-cpp'
        ),
        'dir_pw_third_party_fuzztest': lambda ctx: '"{}"'.format(
            ctx.package_root / 'fuzztest'
        ),
        'dir_pw_third_party_googletest': lambda ctx: '"{}"'.format(
            ctx.package_root / 'googletest'
        ),
        'dir_pw_third_party_re2': lambda ctx: '"{}"'.format(
            ctx.package_root / 're2'
        ),
        'pw_toolchain_OSS_FUZZ_ENABLED': True,
    },
    ninja_targets=('oss_fuzz',),
)


def _env_with_zephyr_vars(ctx: PresubmitContext) -> dict:
    """Returns the environment variables with ... set for Zephyr."""
    env = os.environ.copy()
    # Set some variables here.
    env['ZEPHYR_BASE'] = str(ctx.package_root / 'zephyr')
    env['ZEPHYR_MODULES'] = str(ctx.root)
    env['ZEPHYR_TOOLCHAIN_VARIANT'] = 'llvm'
    return env


def zephyr_build(ctx: PresubmitContext) -> None:
    """Run Zephyr compatible tests"""
    # Install the Zephyr package
    build.install_package(ctx, 'zephyr')
    # Configure the environment
    env = _env_with_zephyr_vars(ctx)
    # Get the python twister runner
    twister = ctx.package_root / 'zephyr' / 'scripts' / 'twister'
    # Get a list of the test roots
    testsuite_roots = [
        ctx.pw_root / dir
        for dir in os.listdir(ctx.pw_root)
        if dir.startswith('pw_')
    ]
    testsuite_roots_list = [
        args for dir in testsuite_roots for args in ('--testsuite-root', dir)
    ]
    sysroot_dir = (
        ctx.pw_root
        / 'environment'
        / 'cipd'
        / 'packages'
        / 'pigweed'
        / 'clang_sysroot'
    )
    platform_filters = (
        ['-P', 'native_posix', '-P', 'native_sim']
        if platform.system() in ['Windows', 'Darwin']
        else []
    )
    # Run twister
    call(
        sys.executable,
        twister,
        '--ninja',
        '--integration',
        '--clobber-output',
        '--inline-logs',
        '--verbose',
        *platform_filters,
        '-x=CONFIG_LLVM_USE_LLD=y',
        '-x=CONFIG_COMPILER_RT_RTLIB=y',
        f'-x=TOOLCHAIN_C_FLAGS=--sysroot={sysroot_dir}',
        f'-x=TOOLCHAIN_LD_FLAGS=--sysroot={sysroot_dir}',
        *testsuite_roots_list,
        env=env,
    )
    # Produces reports at (ctx.root / 'twister_out' / 'twister*.xml')


def docs_build(ctx: PresubmitContext) -> None:
    """Build Pigweed docs"""

    build.install_package(ctx, 'emboss')
    build.install_package(ctx, 'freertos')
    build.install_package(ctx, 'nanopb')
    build.install_package(ctx, 'pico_sdk')
    build.install_package(ctx, 'pigweed_examples_repo')
    build.install_package(ctx, 'stm32cube_f4')

    # Build main docs through GN/Ninja.
    build.gn_gen(ctx, pw_C_OPTIMIZATION_LEVELS=_OPTIMIZATION_LEVELS)
    build.ninja(ctx, 'docs')
    build.gn_check(ctx)

    # Build Rust docs through Bazel.
    build_bazel(
        ctx,
        'build',
        '--',
        '//pw_rust:docs',
    )

    # Build examples repo docs through GN.
    examples_repo_root = ctx.package_root / 'pigweed_examples_repo'
    examples_repo_out = examples_repo_root / 'out'

    # Setup an examples repo presubmit context.
    examples_ctx = PresubmitContext(
        root=examples_repo_root,
        repos=(examples_repo_root,),
        output_dir=examples_repo_out,
        failure_summary_log=ctx.failure_summary_log,
        paths=tuple(),
        all_paths=tuple(),
        package_root=ctx.package_root,
        luci=None,
        override_gn_args={},
        num_jobs=ctx.num_jobs,
        continue_after_build_error=True,
        _failed=False,
        format_options=ctx.format_options,
    )

    # Write a pigweed_environment.gni for the examples repo.
    pwenvgni = (
        ctx.root / 'build_overrides/pigweed_environment.gni'
    ).read_text()
    # Fix the path for cipd packages.
    pwenvgni.replace('../environment/cipd/', '../../cipd/')
    # Write the file
    (examples_repo_root / 'build_overrides/pigweed_environment.gni').write_text(
        pwenvgni
    )

    # Set required GN args.
    pico_sdk_dir = ctx.package_root / 'pico_sdk'
    stm32cube_dir = ctx.package_root / 'stm32cube_f4'
    freertos_dir = ctx.package_root / 'freertos'
    nanopb_dir = ctx.package_root / 'nanopb'
    build.gn_gen(
        examples_ctx,
        dir_pigweed='"//../../.."',
        dir_pw_third_party_stm32cube_f4=f'"{stm32cube_dir}"',
        dir_pw_third_party_freertos=f'"{freertos_dir}"',
        dir_pw_third_party_nanopb=f'"{nanopb_dir}"',
        PICO_SRC_DIR=f'"{pico_sdk_dir}"',
    )
    build.ninja(examples_ctx, 'docs')

    # Copy rust docs from Bazel's out directory into where the GN build
    # put the main docs.
    rust_docs_bazel_dir = ctx.output_dir / 'bazel-bin/pw_rust/docs.rustdoc'
    rust_docs_output_dir = ctx.output_dir / 'docs/gen/docs/html/rustdoc'

    # Copy the doxygen html output to the main docs location.
    doxygen_html_gn_dir = ctx.output_dir / 'docs/doxygen/html'
    doxygen_html_output_dir = ctx.output_dir / 'docs/gen/docs/html/doxygen'

    # Copy the examples repo html output to the main docs location into
    # '/examples/'.
    examples_html_gn_dir = examples_repo_out / 'docs/gen/docs/html'
    examples_html_output_dir = ctx.output_dir / 'docs/gen/docs/html/examples'

    # Remove outputs to avoid including stale files from previous runs.
    shutil.rmtree(rust_docs_output_dir, ignore_errors=True)
    shutil.rmtree(doxygen_html_output_dir, ignore_errors=True)
    shutil.rmtree(examples_html_output_dir, ignore_errors=True)

    # Bazel generates files and directories without write permissions.  In
    # order to allow this rule to be run multiple times we use shutil.copyfile
    # for the actual copies to not copy permissions of files.
    shutil.copytree(
        rust_docs_bazel_dir,
        rust_docs_output_dir,
        copy_function=shutil.copyfile,
        dirs_exist_ok=True,
    )

    # Copy doxygen html outputs.
    shutil.copytree(
        doxygen_html_gn_dir,
        doxygen_html_output_dir,
        copy_function=shutil.copyfile,
        dirs_exist_ok=True,
    )

    # mkdir -p the example repo output dir and copy the files over.
    examples_html_output_dir.mkdir(parents=True, exist_ok=True)
    shutil.copytree(
        examples_html_gn_dir,
        examples_html_output_dir,
        copy_function=shutil.copyfile,
        dirs_exist_ok=True,
    )


gn_host_tools = PigweedGnGenNinja(
    name='gn_host_tools',
    ninja_targets=('host_tools',),
)


def _run_cmake(ctx: PresubmitContext, toolchain='host_clang') -> None:
    build.install_package(ctx, 'nanopb')
    build.install_package(ctx, 'emboss')

    env = None
    if 'clang' in toolchain:
        env = build.env_with_clang_vars()

    toolchain_path = ctx.root / 'pw_toolchain' / toolchain / 'toolchain.cmake'
    build.cmake(
        ctx,
        '--fresh',
        f'-DCMAKE_TOOLCHAIN_FILE={toolchain_path}',
        '-DCMAKE_EXPORT_COMPILE_COMMANDS=1',
        f'-Ddir_pw_third_party_nanopb={ctx.package_root / "nanopb"}',
        '-Dpw_third_party_nanopb_ADD_SUBDIRECTORY=ON',
        f'-Ddir_pw_third_party_emboss={ctx.package_root / "emboss"}',
        env=env,
    )


CMAKE_TARGETS = [
    'pw_apps',
    'pw_run_tests.modules',
]


@filter_paths(
    endswith=(*format_code.C_FORMAT.extensions, '.cmake', 'CMakeLists.txt')
)
def cmake_clang(ctx: PresubmitContext):
    _run_cmake(ctx, toolchain='host_clang')
    build.ninja(ctx, *CMAKE_TARGETS)
    build.gn_check(ctx)


@filter_paths(
    endswith=(*format_code.C_FORMAT.extensions, '.cmake', 'CMakeLists.txt')
)
def cmake_gcc(ctx: PresubmitContext):
    _run_cmake(ctx, toolchain='host_gcc')
    build.ninja(ctx, *CMAKE_TARGETS)
    build.gn_check(ctx)


@filter_paths(
    endswith=(*format_code.C_FORMAT.extensions, '.bazel', '.bzl', 'BUILD')
)
def bazel_test(ctx: PresubmitContext) -> None:
    """Runs bazel test on the entire repo."""
    build_bazel(
        ctx,
        'test',
        '--config=cxx20',
        '--',
        '//...',
    )

    # Run tests for non-default config options

    # pw_rpc
    build_bazel(
        ctx,
        'test',
        '--//pw_rpc:config_override='
        '//pw_rpc:completion_request_callback_config_enabled',
        '--',
        '//pw_rpc/...',
    )

    # pw_grpc
    build_bazel(
        ctx,
        'test',
        '--//pw_rpc:config_override=//pw_grpc:pw_rpc_config',
        '--',
        '//pw_grpc/...',
    )


def bthost_package(ctx: PresubmitContext) -> None:
    """Builds, tests, and prepares bt_host for upload."""
    target = '//pw_bluetooth_sapphire/fuchsia:infra'
    build_bazel(ctx, 'build', '--config=fuchsia', target)

    # Explicitly specify TEST_UNDECLARED_OUTPUTS_DIR_OVERRIDE as that will allow
    # `orchestrate`'s output (eg: ffx host + target logs, test stdout/stderr) to
    # be picked up by the `save_logs` recipe module.
    # We cannot rely on Bazel's native TEST_UNDECLARED_OUTPUTS_DIR functionality
    # since `zip` is not available in builders. See https://pwbug.dev/362990622.
    build_bazel(
        ctx,
        'run',
        '--config=fuchsia',
        f'{target}.test_all',
        env=dict(
            os.environ,
            TEST_UNDECLARED_OUTPUTS_DIR_OVERRIDE=ctx.output_dir,
        ),
    )

    stdout_path = ctx.output_dir / 'bazel.manifest.stdout'
    with open(stdout_path, 'w') as outs:
        build_bazel(
            ctx,
            'build',
            '--config=fuchsia',
            '--output_groups=builder_manifest',
            target,
            stdout=outs,
        )

    manifest_path: Path | None = None
    for line in stdout_path.read_text().splitlines():
        line = line.strip()
        if line.endswith('infrabuilder_manifest.json'):
            manifest_path = Path(line)
            break
    else:
        raise PresubmitFailure('no manifest found in output')

    _LOG.debug('manifest: %s', manifest_path)
    shutil.copyfile(manifest_path, ctx.output_dir / 'builder_manifest.json')


@filter_paths(
    endswith=(
        *format_code.C_FORMAT.extensions,
        '.bazel',
        '.bzl',
        '.py',
        '.rs',
        'BUILD',
    )
)
def bazel_build(ctx: PresubmitContext) -> None:
    """Runs Bazel build for each supported platform."""
    # Build everything with the default flags.
    build_bazel(
        ctx,
        'build',
        '--',
        '//...',
    )

    # Mapping from Bazel platforms to targets which should be built for those
    # platforms.
    targets_for_config = {
        "lm3s6965evb": [
            "//pw_rust/...",
        ],
        "microbit": [
            "//pw_rust/...",
        ],
    }

    for cxxversion in ('17', '20'):
        # Explicitly build for each supported C++ version.
        args = [ctx, 'build', f"--//pw_toolchain/cc:cxx_standard={cxxversion}"]
        args += ['--', '//...']
        build_bazel(*args)

        for config, targets in targets_for_config.items():
            build_bazel(
                ctx,
                'build',
                f'--config={config}',
                f"--//pw_toolchain/cc:cxx_standard={cxxversion}",
                *targets,
            )

    build_bazel(
        ctx,
        'build',
        '--config=stm32f429i_freertos',
        '--//pw_thread_freertos:config_override=//pw_build:test_module_config',
        '//pw_build:module_config_test',
    )

    # Build upstream Pigweed for the rp2040.
    # First using the config.
    build_bazel(
        ctx,
        'build',
        '--config=rp2040',
        '//...',
        # Bazel will silently skip any incompatible targets in wildcard builds;
        # but we know that some end-to-end targets definitely should remain
        # compatible with this platform. So we list them explicitly. (If an
        # explicitly listed target is incompatible with the platform, Bazel
        # will return an error instead of skipping it.)
        '//pw_system:system_example',
    )
    # Then using the transition.
    #
    # This ensures that the rp2040_binary rule transition includes all required
    # backends.
    build_bazel(
        ctx,
        'build',
        '//pw_system:rp2040_system_example',
    )

    # Build upstream Pigweed for the Discovery board using STM32Cube.
    build_bazel(
        ctx,
        'build',
        '--config=stm32f429i_freertos',
        '//...',
        # Bazel will silently skip any incompatible targets in wildcard builds;
        # but we know that some end-to-end targets definitely should remain
        # compatible with this platform. So we list them explicitly. (If an
        # explicitly listed target is incompatible with the platform, Bazel
        # will return an error instead of skipping it.)
        '//pw_system:system_example',
    )

    # Build upstream Pigweed for the Discovery board using the baremetal
    # backends.
    build_bazel(
        ctx,
        'build',
        '--config=stm32f429i_baremetal',
        '//...',
    )

    # Build the fuzztest example.
    #
    # TODO: b/324652164 - This doesn't work on MacOS yet.
    if sys.platform != 'darwin':
        build_bazel(
            ctx,
            'build',
            '--config=fuzztest',
            '//pw_fuzzer/examples/fuzztest:metrics_fuzztest',
        )


def pw_transfer_integration_test(ctx: PresubmitContext) -> None:
    """Runs the pw_transfer cross-language integration test only.

    This test is not part of the regular bazel build because it's slow and
    intended to run in CI only.
    """
    build_bazel(
        ctx,
        'test',
        '//pw_transfer/integration_test:cross_language_small_test',
        '//pw_transfer/integration_test:cross_language_medium_read_test',
        '//pw_transfer/integration_test:cross_language_medium_write_test',
        '//pw_transfer/integration_test:cross_language_large_read_test',
        '//pw_transfer/integration_test:cross_language_large_write_test',
        '//pw_transfer/integration_test:multi_transfer_test',
        '//pw_transfer/integration_test:expected_errors_test',
        '//pw_transfer/integration_test:legacy_binaries_test',
        '--test_output=errors',
    )


#
# General presubmit checks
#


def _clang_system_include_paths(lang: str) -> list[str]:
    """Generate default system header paths.

    Returns the list of system include paths used by the host
    clang installation.
    """
    # Dump system include paths with preprocessor verbose.
    command = [
        'clang++',
        '-Xpreprocessor',
        '-v',
        '-x',
        f'{lang}',
        f'{os.devnull}',
        '-fsyntax-only',
    ]
    process = log_run(
        command, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )

    # Parse the command output to retrieve system include paths.
    # The paths are listed one per line.
    output = process.stdout.decode(errors='backslashreplace')
    include_paths: list[str] = []
    for line in output.splitlines():
        path = line.strip()
        if os.path.exists(path):
            include_paths.append(f'-isystem{path}')

    return include_paths


def edit_compile_commands(
    in_path: Path, out_path: Path, func: Callable[[str, str, str], str]
) -> None:
    """Edit the selected compile command file.

    Calls the input callback on all triplets (file, directory, command) in
    the input compile commands database. The return value replaces the old
    compile command in the output database.
    """
    with open(in_path) as in_file:
        compile_commands = json.load(in_file)
        for item in compile_commands:
            item['command'] = func(
                item['file'], item['directory'], item['command']
            )
    with open(out_path, 'w') as out_file:
        json.dump(compile_commands, out_file, indent=2)


_EXCLUDE_FROM_COPYRIGHT_NOTICE: Sequence[str] = (
    # Configuration
    # keep-sorted: start
    r'MODULE.bazel.lock',
    r'\b49-pico.rules$',
    r'\bDoxyfile$',
    r'\bPW_PLUGINS$',
    r'\bconstraint.list$',
    r'\bconstraint_hashes_darwin.list$',
    r'\bconstraint_hashes_linux.list$',
    r'\bconstraint_hashes_windows.list$',
    r'\bpython_base_requirements.txt$',
    r'\bupstream_requirements_darwin_lock.txt$',
    r'\bupstream_requirements_linux_lock.txt$',
    r'\bupstream_requirements_windows_lock.txt$',
    r'^(?:.+/)?\.bazelversion$',
    r'^pw_env_setup/py/pw_env_setup/cipd_setup/.cipd_version',
    # keep-sorted: end
    # Metadata
    # keep-sorted: start
    r'\b.*OWNERS.*$',
    r'\bAUTHORS$',
    r'\bLICENSE$',
    r'\bPIGWEED_MODULES$',
    r'\b\.vscodeignore$',
    r'\bgo.(mod|sum)$',
    r'\bpackage-lock.json$',
    r'\bpackage.json$',
    r'\brequirements.txt$',
    r'\byarn.lock$',
    r'^docker/tag$',
    r'^patches.json$',
    # keep-sorted: end
    # Data files
    # keep-sorted: start
    r'\.bin$',
    r'\.csv$',
    r'\.elf$',
    r'\.gif$',
    r'\.ico$',
    r'\.jpg$',
    r'\.json$',
    r'\.png$',
    r'\.svg$',
    r'\.vsix$',
    r'\.woff2',
    r'\.xml$',
    # keep-sorted: end
    # Documentation
    # keep-sorted: start
    r'\.md$',
    r'\.rst$',
    # keep-sorted: end
    # Generated protobuf files
    # keep-sorted: start
    r'\.pb\.c$',
    r'\.pb\.h$',
    r'\_pb2.pyi?$',
    # keep-sorted: end
    # Generated third-party files
    # keep-sorted: start
    r'\bthird_party/.*\.bazelrc$',
    r'\bthird_party/fuchsia/repo',
    r'\bthird_party/perfetto/repo/protos/perfetto/trace/perfetto_trace.proto',
    # keep-sorted: end
    # Diff/Patch files
    # keep-sorted: start
    r'\.diff$',
    r'\.patch$',
    # keep-sorted: end
    # Test data
    # keep-sorted: start
    r'\bpw_presubmit/py/test/owners_checks/',
    # keep-sorted: end
)

# Regular expression for the copyright comment. "\1" refers to the comment
# characters and "\2" refers to space after the comment characters, if any.
# All period characters are escaped using a replace call.
# pylint: disable=line-too-long
_COPYRIGHT = re.compile(
    r"""(#|//|::| \*|)( ?)Copyright 2\d{3} The Pigweed Authors
\1
\1\2Licensed under the Apache License, Version 2.0 \(the "License"\); you may not
\1\2use this file except in compliance with the License. You may obtain a copy of
\1\2the License at
\1
\1(?:\2    |\t)https://www.apache.org/licenses/LICENSE-2.0
\1
\1\2Unless required by applicable law or agreed to in writing, software
\1\2distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
\1\2WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
\1\2License for the specific language governing permissions and limitations under
\1\2the License.
""".replace(
        '.', r'\.'
    ),
    re.MULTILINE,
)
# pylint: enable=line-too-long

_SKIP_LINE_PREFIXES = (
    '#!',
    '#autoload',
    '#compdef',
    '@echo off',
    ':<<',
    '/*',
    ' * @jest-environment jsdom',
    ' */',
    '{#',  # Jinja comment block
    '# -*- coding: utf-8 -*-',
    '<!--',
)


def _read_notice_lines(file: TextIO) -> Iterable[str]:
    lines = iter(file)
    try:
        # Read until the first line of the copyright notice.
        line = next(lines)
        while line.isspace() or line.startswith(_SKIP_LINE_PREFIXES):
            line = next(lines)

        yield line

        for _ in range(12):  # The notice is 13 lines; read the remaining 12.
            yield next(lines)
    except StopIteration:
        return


@filter_paths(exclude=_EXCLUDE_FROM_COPYRIGHT_NOTICE)
def copyright_notice(ctx: PresubmitContext):
    """Checks that the Pigweed copyright notice is present."""
    errors = []

    for path in ctx.paths:
        if path.stat().st_size == 0:
            continue  # Skip empty files

        try:
            with path.open() as file:
                if not _COPYRIGHT.match(''.join(_read_notice_lines(file))):
                    errors.append(path)
        except UnicodeDecodeError as exc:
            raise PresubmitFailure(f'failed to read {path}') from exc

    if errors:
        _LOG.warning(
            '%s with a missing or incorrect copyright notice:\n%s',
            plural(errors, 'file'),
            '\n'.join(str(e) for e in errors),
        )
        raise PresubmitFailure


@filter_paths(endswith=format_code.CPP_SOURCE_EXTS)
def source_is_in_cmake_build_warn_only(ctx: PresubmitContext):
    """Checks that source files are in the CMake build."""

    _run_cmake(ctx)
    missing = SOURCE_FILES_FILTER_CMAKE_EXCLUDE.filter(
        build.check_compile_commands_for_files(
            ctx.output_dir / 'compile_commands.json',
            (f for f in ctx.paths if f.suffix in format_code.CPP_SOURCE_EXTS),
        )
    )
    if missing:
        _LOG.warning(
            'Files missing from CMake:\n%s',
            '\n'.join(str(f) for f in missing),
        )


def build_env_setup(ctx: PresubmitContext):
    if 'PW_CARGO_SETUP' not in os.environ:
        _LOG.warning('Skipping build_env_setup since PW_CARGO_SETUP is not set')
        return

    tmpl = ctx.root.joinpath('pw_env_setup', 'py', 'pyoxidizer.bzl.tmpl')
    out = ctx.output_dir.joinpath('pyoxidizer.bzl')

    with open(tmpl, 'r') as ins:
        cfg = ins.read().replace('${PW_ROOT}', str(ctx.root))
        with open(out, 'w') as outs:
            outs.write(cfg)

    call('pyoxidizer', 'build', cwd=ctx.output_dir)


def _valid_capitalization(word: str) -> bool:
    """Checks that the word has a capital letter or is not a regular word."""
    return bool(
        any(c.isupper() for c in word)  # Any capitalizatian (iTelephone)
        or not word.isalpha()  # Non-alphabetical (cool_stuff.exe)
        or shutil.which(word)
    )  # Matches an executable (clangd)


def commit_message_format(_: PresubmitContext):
    """Checks that the top commit's message is correctly formatted."""
    if git_repo.commit_author().endswith('gserviceaccount.com'):
        return

    lines = git_repo.commit_message().splitlines()

    # Show limits and current commit message in log.
    _LOG.debug('%-25s%+25s%+22s', 'Line limits', '72|', '72|')
    for line in lines:
        _LOG.debug(line)

    if not lines:
        _LOG.error('The commit message is too short!')
        raise PresubmitFailure

    # Ignore merges.
    repo = git_repo.LoggingGitRepo(Path.cwd())
    parents = repo.commit_parents()
    _LOG.debug('parents: %r', parents)
    if len(parents) > 1:
        _LOG.warning('Ignoring multi-parent commit')
        return

    # Ignore Gerrit-generated reverts.
    if (
        'Revert' in lines[0]
        and 'This reverts commit ' in git_repo.commit_message()
        and 'Reason for revert: ' in git_repo.commit_message()
    ):
        _LOG.warning('Ignoring apparent Gerrit-generated revert')
        return

    # Ignore Gerrit-generated relands
    if (
        'Reland' in lines[0]
        and 'This is a reland of ' in git_repo.commit_message()
        and "Original change's description:" in git_repo.commit_message()
    ):
        _LOG.warning('Ignoring apparent Gerrit-generated reland')
        return

    errors = 0

    if len(lines[0]) > 72:
        _LOG.warning(
            "The commit message's first line must be no longer than "
            '72 characters.'
        )
        _LOG.warning(
            'The first line is %d characters:\n  %s', len(lines[0]), lines[0]
        )
        errors += 1

    if lines[0].endswith('.'):
        _LOG.warning(
            "The commit message's first line must not end with a period:\n %s",
            lines[0],
        )
        errors += 1

    # Check that the first line matches the expected pattern.
    match = re.match(
        r'^(?:[.\w*/]+(?:{[\w* ,]+})?[\w*/]*|SEED-\d+): (?P<desc>.+)$', lines[0]
    )
    if not match:
        _LOG.warning('The first line does not match the expected format')
        _LOG.warning(
            'Expected:\n\n  module_or_target: The description\n\n'
            'Found:\n\n  %s\n',
            lines[0],
        )
        errors += 1
    elif not _valid_capitalization(match.group('desc').split()[0]):
        _LOG.warning(
            'The first word after the ":" in the first line ("%s") must be '
            'capitalized:\n  %s',
            match.group('desc').split()[0],
            lines[0],
        )
        errors += 1

    if len(lines) > 1 and lines[1]:
        _LOG.warning("The commit message's second line must be blank.")
        _LOG.warning(
            'The second line has %d characters:\n  %s', len(lines[1]), lines[1]
        )
        errors += 1

    # Ignore the line length check for Copybara imports so they can include the
    # commit hash and description for imported commits.
    if not errors and (
        'Copybara import' in lines[0]
        and 'GitOrigin-RevId:' in git_repo.commit_message()
    ):
        _LOG.warning('Ignoring Copybara import')
        return

    # Check that the lines are 72 characters or less.
    for i, line in enumerate(lines[2:], 3):
        # Skip any lines that might possibly have a URL, path, or metadata in
        # them.
        if any(c in line for c in ':/>'):
            continue

        # Skip any lines with non-ASCII characters.
        if not line.isascii():
            continue

        # Skip any blockquoted lines.
        if line.startswith('  '):
            continue

        if len(line) > 72:
            _LOG.warning(
                'Commit message lines must be no longer than 72 characters.'
            )
            _LOG.warning('Line %d has %d characters:\n  %s', i, len(line), line)
            errors += 1

    if errors:
        _LOG.error('Found %s in the commit message', plural(errors, 'error'))
        raise PresubmitFailure


@filter_paths(endswith=(*format_code.C_FORMAT.extensions, '.py'))
def static_analysis(ctx: PresubmitContext):
    """Runs all available static analysis tools."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'python.lint', 'static_analysis')
    build.gn_check(ctx)


_EXCLUDE_FROM_TODO_CHECK = (
    # keep-sorted: start
    r'.bazelrc$',
    r'.dockerignore$',
    r'.gitignore$',
    r'.pylintrc$',
    r'\bdocs/build_system.rst',
    r'\bdocs/code_reviews.rst',
    r'\bpw_assert_basic/basic_handler.cc',
    r'\bpw_assert_basic/public/pw_assert_basic/handler.h',
    r'\bpw_blob_store/public/pw_blob_store/flat_file_system_entry.h',
    r'\bpw_build/linker_script.gni',
    r'\bpw_build/py/pw_build/copy_from_cipd.py',
    r'\bpw_cpu_exception/basic_handler.cc',
    r'\bpw_cpu_exception_cortex_m/entry.cc',
    r'\bpw_cpu_exception_cortex_m/exception_entry_test.cc',
    r'\bpw_doctor/py/pw_doctor/doctor.py',
    r'\bpw_env_setup/util.sh',
    r'\bpw_fuzzer/fuzzer.gni',
    r'\bpw_i2c/BUILD.gn',
    r'\bpw_i2c/public/pw_i2c/register_device.h',
    r'\bpw_kvs/flash_memory.cc',
    r'\bpw_kvs/key_value_store.cc',
    r'\bpw_log_basic/log_basic.cc',
    r'\bpw_package/py/pw_package/packages/chromium_verifier.py',
    r'\bpw_protobuf/encoder.cc',
    r'\bpw_rpc/docs.rst',
    r'\bpw_watch/py/pw_watch/watch.py',
    r'\btargets/mimxrt595_evk/BUILD.bazel',
    r'\btargets/stm32f429i_disc1/boot.cc',
    r'\bthird_party/chromium_verifier/BUILD.gn',
    # keep-sorted: end
)


@filter_paths(exclude=_EXCLUDE_FROM_TODO_CHECK)
def todo_check_with_exceptions(ctx: PresubmitContext):
    """Check that non-legacy TODO lines are valid."""  # todo-check: ignore
    todo_check.create(todo_check.BUGS_OR_USERNAMES)(ctx)


@filter_paths(file_filter=format_code.OWNERS_CODE_FORMAT.filter)
def owners_lint_checks(ctx: PresubmitContext):
    """Runs OWNERS linter."""
    owners_checks.presubmit_check(ctx.paths)


SOURCE_FILES_FILTER = FileFilter(
    endswith=_BUILD_FILE_FILTER.endswith,
    suffix=('.bazel', '.bzl', '.gn', '.gni', *_BUILD_FILE_FILTER.suffix),
    exclude=(
        r'zephyr.*',
        r'android.*',
        r'\.black.toml',
        r'pyproject.toml',
    ),
)

SOURCE_FILES_FILTER_GN_EXCLUDE = FileFilter(
    exclude=(
        # keep-sorted: start
        r'\bpw_bluetooth_sapphire/fuchsia',
        # keep-sorted: end
    ),
)

SOURCE_FILES_FILTER_CMAKE_EXCLUDE = FileFilter(
    exclude=(
        # keep-sorted: start
        r'\bpw_bluetooth_sapphire/fuchsia',
        # keep-sorted: end
    ),
)

#
# Presubmit check programs
#

OTHER_CHECKS = (
    # keep-sorted: start
    bazel_test,
    bthost_package,
    build.gn_gen_check,
    cmake_clang,
    cmake_gcc,
    coverage,
    # TODO: b/234876100 - Remove once msan is added to all_sanitizers().
    cpp_checks.msan,
    docs_build,
    gitmodules.create(gitmodules.Config(allow_submodules=False)),
    gn_all,
    gn_clang_build,
    gn_combined_build_check,
    gn_main_build_check,
    gn_platform_build_check,
    module_owners.presubmit_check(),
    npm_presubmit.npm_test,
    pw_transfer_integration_test,
    python_checks.update_upstream_python_constraints,
    python_checks.vendor_python_wheels,
    python_checks.version_bump_pigweed_pypi_distribution,
    python_checks.upload_pigweed_pypi_distribution,
    shell_checks.shellcheck,
    # TODO(hepler): Many files are missing from the CMake build. Add this check
    # to lintformat when the missing files are fixed.
    source_in_build.cmake(SOURCE_FILES_FILTER, _run_cmake),
    static_analysis,
    stm32f429i,
    todo_check.create(todo_check.BUGS_OR_USERNAMES),
    zephyr_build,
    # keep-sorted: end
)

ARDUINO_PICO = (
    # Skip gn_teensy_build if running on mac-arm64.
    # There are no arm specific tools packages available upstream:
    # https://www.pjrc.com/teensy/package_teensy_index.json
    gn_teensy_build
    if not (sys.platform == 'darwin' and platform.machine() == 'arm64')
    else (),
    gn_pico_build,
    gn_pw_system_demo_build,
)

INTERNAL = (gn_mimxrt595_build, gn_mimxrt595_freertos_build)

SAPPHIRE = (
    # keep-sorted: start
    gn_chre_googletest_nanopb_sapphire_build,
    # keep-sorted: end
)

SANITIZERS = (cpp_checks.all_sanitizers(),)

SECURITY = (
    # keep-sorted: start
    gn_crypto_mbedtls_build,
    gn_crypto_micro_ecc_build,
    gn_software_update_build,
    # keep-sorted: end
)

FUZZ = (gn_fuzz_build, oss_fuzz_build)

_LINTFORMAT = (
    commit_message_format,
    copyright_notice,
    format_code.presubmit_checks(),
    inclusive_language.presubmit_check.with_filter(
        exclude=(
            r'\bMODULE.bazel.lock$',
            r'\bgo.sum$',
            r'\bpackage-lock.json$',
            r'\byarn.lock$',
        )
    ),
    cpp_checks.pragma_once,
    build.bazel_lint,
    owners_lint_checks,
    source_in_build.gn(SOURCE_FILES_FILTER).with_file_filter(
        SOURCE_FILES_FILTER_GN_EXCLUDE
    ),
    source_is_in_cmake_build_warn_only,
    javascript_checks.eslint if shutil.which('npm') else (),
    json_check.presubmit_check,
    keep_sorted.presubmit_check,
    todo_check_with_exceptions,
)

LINTFORMAT = (
    _LINTFORMAT,
    # This check is excluded from _LINTFORMAT because it's not quick: it issues
    # a bazel query that pulls in all of Pigweed's external dependencies
    # (https://stackoverflow.com/q/71024130/1224002). These are cached, but
    # after a roll it can be quite slow.
    source_in_build.bazel(SOURCE_FILES_FILTER),
    python_checks.check_python_versions,
    python_checks.gn_python_lint,
)

QUICK = (
    _LINTFORMAT,
    gn_quick_build_check,
)

FULL = (
    _LINTFORMAT,
    gn_combined_build_check,
    gn_host_tools,
    bazel_test,
    bazel_build,
    python_checks.gn_python_check,
    python_checks.gn_python_test_coverage,
    python_checks.check_upstream_python_constraints,
    build_env_setup,
)

PROGRAMS = Programs(
    # keep-sorted: start
    arduino_pico=ARDUINO_PICO,
    full=FULL,
    fuzz=FUZZ,
    internal=INTERNAL,
    lintformat=LINTFORMAT,
    other_checks=OTHER_CHECKS,
    quick=QUICK,
    sanitizers=SANITIZERS,
    sapphire=SAPPHIRE,
    security=SECURITY,
    # keep-sorted: end
)


def parse_args() -> argparse.Namespace:
    """Creates an argument parser and parses arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    cli.add_arguments(parser, PROGRAMS, 'quick')
    parser.add_argument(
        '--install',
        action='store_true',
        help='Install the presubmit as a Git pre-push hook and exit.',
    )

    return parser.parse_args()


def run(install: bool, **presubmit_args) -> int:
    """Entry point for presubmit."""

    if install:
        install_git_hook(
            'pre-push',
            [
                'python',
                '-m',
                'pw_presubmit.pigweed_presubmit',
                '--base',
                'origin/main..HEAD',
                '--program',
                'quick',
            ],
        )
        return 0

    return cli.run(**presubmit_args)


def main() -> int:
    """Run the presubmit for the Pigweed repository."""
    return run(**vars(parse_args()))


if __name__ == '__main__':
    try:
        # If pw_cli is available, use it to initialize logs.
        from pw_cli import log  # pylint: disable=ungrouped-imports

        log.install(logging.INFO)
    except ImportError:
        # If pw_cli isn't available, display log messages like a simple print.
        logging.basicConfig(format='%(message)s', level=logging.INFO)

    sys.exit(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_cpu_exception_cortex_m_protos/cpu_state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0pw_cpu_exception_cortex_m_protos/cpu_state.proto\x12\x19pw.cpu_exception.cortex_m\"\x89\x03\n\x0e\x41rmV7mCpuState\x12\n\n\x02pc\x18\x01 \x01(\r\x12\n\n\x02lr\x18\x02 \x01(\r\x12\x0b\n\x03psr\x18\x03 \x01(\r\x12\x0b\n\x03msp\x18\x04 \x01(\r\x12\x0b\n\x03psp\x18\x05 \x01(\r\x12\x12\n\nexc_return\x18\x06 \x01(\r\x12\x0c\n\x04\x63\x66sr\x18\x07 \x01(\r\x12\x0e\n\x06msplim\x18\x1b \x01(\r\x12\x0e\n\x06psplim\x18\x1c \x01(\r\x12\r\n\x05mmfar\x18\x08 \x01(\r\x12\x0c\n\x04\x62\x66\x61r\x18\t \x01(\r\x12\x0c\n\x04icsr\x18\n \x01(\r\x12\x0c\n\x04hfsr\x18\x19 \x01(\r\x12\r\n\x05shcsr\x18\x1a \x01(\r\x12\x0f\n\x07\x63ontrol\x18\x0b \x01(\r\x12\n\n\x02r0\x18\x0c \x01(\r\x12\n\n\x02r1\x18\r \x01(\r\x12\n\n\x02r2\x18\x0e \x01(\r\x12\n\n\x02r3\x18\x0f \x01(\r\x12\n\n\x02r4\x18\x10 \x01(\r\x12\n\n\x02r5\x18\x11 \x01(\r\x12\n\n\x02r6\x18\x12 \x01(\r\x12\n\n\x02r7\x18\x13 \x01(\r\x12\n\n\x02r8\x18\x14 \x01(\r\x12\n\n\x02r9\x18\x15 \x01(\r\x12\x0b\n\x03r10\x18\x16 \x01(\r\x12\x0b\n\x03r11\x18\x17 \x01(\r\x12\x0b\n\x03r12\x18\x18 \x01(\r\"^\n\x17SnapshotCpuStateOverlay\x12\x43\n\x10\x61rmv7m_cpu_state\x18\x14 \x01(\x0b\x32).pw.cpu_exception.cortex_m.ArmV7mCpuState')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_cpu_exception_cortex_m_protos.cpu_state_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_ARMV7MCPUSTATE']._serialized_start=80
  _globals['_ARMV7MCPUSTATE']._serialized_end=473
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_start=475
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_end=569
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""GN output support."""

from typing import Any

import pathlib

try:
    from pw_build_mcuxpresso.components import Project
except ImportError:
    # Load from this directory if pw_build_mcuxpresso is not available.
    from components import Project  # type: ignore


def _gn_list_str_out(name: str, val: list[Any]):
    """Outputs list of strings in GN format with correct escaping."""
    list_str = ','.join(
        '"' + str(x).replace('"', r'\"').replace('$', r'\$') + '"' for x in val
    )
    print(f'{name} = [{list_str}]')


def _gn_list_path_out(
    name: str, val: list[pathlib.Path], path_prefix: str | None = None
):
    """Outputs list of paths in GN format with common prefix."""
    if path_prefix is not None:
        str_val = list(f'{path_prefix}/{str(d)}' for d in val)
    else:
        str_val = list(str(d) for d in val)
    _gn_list_str_out(name, str_val)


def gn_output(project: Project, path_prefix: str | None = None):
    """Output GN scope for a project with the specified components.

    Args:
        project: MCUXpresso project to output..
        path_prefix: string prefix to prepend to all paths.
    """
    _gn_list_str_out('defines', project.defines)
    _gn_list_path_out(
        'include_dirs', project.include_dirs, path_prefix=path_prefix
    )
    _gn_list_path_out('public', project.headers, path_prefix=path_prefix)
    _gn_list_path_out('sources', project.sources, path_prefix=path_prefix)
    _gn_list_path_out('libs', project.libs, path_prefix=path_prefix)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Attractive status output to the terminal (and other places if you want)."""

import logging
from typing import Callable

from pw_cli.color import colors


def _no_color(msg: str) -> str:
    return msg


def _split_lines(msg: str | list[str]) -> tuple[str, list[str]]:
    """Turn a list of strings into a tuple of the first and list of rest."""
    if isinstance(msg, str):
        return (msg, [])

    return (msg[0], msg[1:])


class StatusReporter:
    """Print user-friendly status reports to the terminal for CLI tools.

    You can instead redirect these lines to logs without formatting by
    substituting ``LoggingStatusReporter``. Consumers of this should be
    designed to take any subclass and not make assumptions about where the
    output will go. But the reason you would choose this over plain logging is
    because you want to support pretty-printing to the terminal.

    This is also "themable" in the sense that you can subclass this, override
    the methods with whatever formatting you want, and supply the subclass to
    anything that expects an instance of this.

    Key:

    - info: Plain ol' informational status.
    - ok: Something was checked and it was okay.
    - new: Something needed to be changed/updated and it was successfully.
    - wrn: Warning, non-critical.
    - err: Error, critical.

    This doesn't expose the %-style string formatting that is used in idiomatic
    Python logging, but this shouldn't be used for performance-critical logging
    situations anyway.
    """

    def _report(  # pylint: disable=no-self-use
        self,
        msg: str | list[str],
        color: Callable[[str], str],
        char: str,
        func: Callable,
        silent: bool,
    ) -> None:
        """Actually print/log/whatever the status lines."""
        first_line, rest_lines = _split_lines(msg)
        first_line = color(f'{char} {first_line}')
        spaces = ' ' * len(char)
        rest_lines = [color(f'{spaces} {line}') for line in rest_lines]

        if not silent:
            for line in [first_line, *rest_lines]:
                func(line)

    def demo(self):
        """Run this to see what your status reporter output looks like."""
        self.info(
            [
                'FYI, here\'s some information:',
                'Lorem ipsum dolor sit amet, consectetur adipiscing elit.',
                'Donec condimentum metus molestie metus maximus ultricies '
                'ac id dolor.',
            ]
        )
        self.ok('This is okay, no changes needed.')
        self.new('We changed some things successfully!')
        self.wrn('Uh oh, you might want to be aware of this.')
        self.err('This is bad! Things might be broken!')

    def info(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '\u2022', print, silent)

    def ok(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, colors().blue, '\u2713', print, silent)

    def new(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, colors().green, '\u2713', print, silent)

    def wrn(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, colors().yellow, '\u26A0\uFE0F ', print, silent)

    def err(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, colors().red, '\U0001F525', print, silent)


class LoggingStatusReporter(StatusReporter):
    """Print status lines to logs instead of to the terminal."""

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger
        super().__init__()

    def _report(
        self,
        msg: str | list[str],
        color: Callable[[str], str],
        char: str,
        func: Callable,
        silent: bool,
    ) -> None:
        first_line, rest_lines = _split_lines(msg)

        if not silent:
            for line in [first_line, *rest_lines]:
                func(line)

    def info(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def ok(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def new(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.info, silent)

    def wrn(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.warning, silent)

    def err(self, msg: str | list[str], silent: bool = False) -> None:
        self._report(msg, _no_color, '', self.logger.error, silent)
